/****************************************************************
  llpc.pl
	LLP --> LLPAM Compiler
	by Naoyuki Tamura (tamura@kobe-u.ac.jp)
	   Dept. CS, Kobe University
        and 
           Mutsunori Banbara (banbara@cs32.scitec.kobe-u.ac.jp)
           Nara National College of Technology
        Last modified on 10 July 2000 by M.Banbara
****************************************************************/
%:- op(1200, fx,  (forall)).
%:- op(1190, xfx, (\)).
%:- op(1170, xfx, (<=)).
%:- op(1170, xfx, (:-)).
%:- op(1170, xfx, (-->)).
%:- op(1170,  fx, (:-)).
%:- op(1170,  fx, (?-)).
:- op(1200,  fy, (linear)).
:- op(1200,  fy, (forall)).
:- op(1200, xfy, (\)).
:- op(1200, yfx, (<=)).
:- op(1150,  fx, (resource)).
:- op(1060, xfy, (&)).
:- op( 950, xfy, (-<>)).
:- op( 950, xfy, (=>)).
:- op( 900,  fy, (!)).

%%% for SWI-Prolog
:- op(1150,  fx, (mode)).
:- op(1150,  fx, (public)).
:- op(1150,  fx, (dynamic)).
:- op(1150,  fx, (block)).
:- op(1150,  fx, (meta_predicate)).
:- op( 500, yfx, (#)).

:- dynamic internal_clause/2, internal_predicates/2.
:- dynamic compiled_resource/1, resource_predicates/2.
:- dynamic linear_clause/1.
:- dynamic dynamic_predicates/2.
:- dynamic filename/1.
:- dynamic disj_counter/1, res_counter/1.

%%% for Quintus Prolog
runtime_entry(start) :-		
	main.

main :-
	read(X),
	llpc(X).

llpc([LLPFile,AsmFile]) :-
	llpc(LLPFile, AsmFile).
llpc(_).

llpc(LLPFile, AsmFile) :-
	read_file(LLPFile),
	open(AsmFile, write, Out),
	compile_all_predicates(Out),
	close(Out).

/****************************************************************
	Read LLP program
****************************************************************/
read_file(LLPFile) :-
	llpc_init(LLPFile),
	open(LLPFile, read, In),
	repeat,
	  read_stream(In, X),
	  assert_clause(X),
	X == end_of_file,
	close(In).

llpc_init(File) :-
	retractall(internal_clause(_,_)),
	retractall(internal_predicates(_,_)),
	retractall(linear_clause(_)),
	retractall(resource_predicates(_,_)),
	retractall(compiled_resource(_)),
	retractall(disj_counter(_)),
	retractall(res_counter(_)),
	retractall(filename(_)),
	assert(filename(File)),
	assert(disj_counter(0)),
	assert(res_counter(0)).

assert_clause(end_of_file) :- !.
assert_clause((:- resource G)) :- !, 
	conj_to_list(G, G1),
	assert_res_cls(G1).
assert_clause((:- dynamic _G)) :- !.
%%% not implemented yet.
%assert_clause((:- dynamic G)) :- !,
%	conj_to_list(G, G1),
%	assert_dyn_predicate(G1).
assert_clause((:- mode _G)) :- !.
assert_clause((:- multifile _G)) :- !.
assert_clause((:- public _G)) :- !.
assert_clause((:- block _G)) :- !.
assert_clause((:- meta_predicate _G)) :- !.
assert_clause((:- Goal)) :- !,
	call(Goal).
assert_clause((linear Clause)) :- !,
	preprocess_linear_clause(Clause, Clause1),
	asserta(linear_clause(Clause1)).
assert_clause((C1 & C2)) :- !,
	assert_clause(C1),
	assert_clause(C2).
assert_clause(Clause0) :-
	preprocess(Clause0, Clause),
	assert_cls(Clause).

assert_cls((Head :- Body)) :- !,
	assert_predicate(Head),
	assert(internal_clause(Head, Body)).
assert_cls(Head) :- !,
	assert_predicate(Head),
	assert(internal_clause(Head, true)).

assert_predicate(Head) :-
	functor(Head, Functor, Arity),
	clause(internal_predicates(Functor, Arity), _),
	!.
assert_predicate(Head) :-
	functor(Head, Functor, Arity),
	assert(internal_predicates(Functor, Arity)).

assert_res_cls([]) :- !.
assert_res_cls([G|Gs]) :- 
	G = Functor/Arity, 
	!,
	functor(Head, Functor, Arity),
	assert_cls((Head :- call_resource(Head))),
	assert_res_cls(Gs).
assert_res_cls(G) :- 
	llpc_error([G,is,invalid,resource,declaration]),
	fail.

assert_res_predicate([]) :- !.
assert_res_predicate([G|Gs]) :- 
	G = Functor/Arity, 
	!,
	assert_res(Functor, Arity),
	functor(Head, Functor, Arity),
	assert_predicate(Head),
	assert_res_predicate(Gs).
assert_res_predicate(G) :- 
	llpc_error([G,is,invalid,resource,functor]),
	fail.

assert_res(Functor, Arity) :- 
	clause(resource_predicates(Functor, Arity), _), !.
assert_res(Functor, Arity) :- 
	assert(resource_predicates(Functor, Arity)).

assert_dyn_predicate([]) :- !.
assert_dyn_predicate([G|Gs]) :-
	G = Functor/Arity, 
	!,
	assert_dyn(Functor, Arity),
	functor(Head, Functor, Arity),
	assert_cls((Head :- clause(Head, Goal), call(Goal))),
	assert_dyn_predicate(Gs).
assert_dyn_predicate(G) :- 
	llpc_error([G,is,invalid,dynamic,declaration]),
	fail.

assert_dyn(Functor, Arity) :-
	clause(dynamic_predicates(Functor, Arity), _), !.
assert_dyn(Functor, Arity) :-
	assert(dynamic_predicates(Functor, Arity)).

preprocess_linear_clause((C1 & C2), (NewC1 & NewC2)) :- !,
	preprocess_linear_clause(C1, NewC1),
	preprocess_linear_clause(C2, NewC2).
preprocess_linear_clause(C0, (forall Xs \ C)) :-
	preprocess(C0, C1),
	copy_term(C1, C2),
	preprocess_linear_clause1(C2, C),
	variables(C, Xs).

preprocess_linear_clause1((H :- B), (B -<> H)) :- !,
	functor(H, F, A),
	assert_res_predicate([F/A]).
preprocess_linear_clause1(H, H) :-
	functor(H, F, A),
	assert_res_predicate([F/A]).

preprocess(Clause0, Clause) :-
	expand_term(Clause0, Clause1),
	expand_resource(Clause1, Clause2, ResFA),
	assert_res_predicate(ResFA),
	eliminate_disj(Clause2, Clause, DummyClauses),
	assert_dummy_clauses(DummyClauses).

assert_dummy_clauses([]) :- !.
assert_dummy_clauses([C|Cs]) :- 
	assert_clause(C),
	assert_dummy_clauses(Cs).

/****************************************************************
	Compile LLP program
****************************************************************/
compile_all_predicates(Out) :-
	findall(C, linear_clause(C), Clauses),
	compile_linear_clauses(Clauses, Instructions, []),
	write_asm(Out, Instructions),
	nl_stream(Out), 
	fail.
compile_all_predicates(Out) :-
	clause(internal_predicates(Functor, Arity), _),
	compile_predicate(Functor, Arity, Instructions, []),
	write_asm(Out, Instructions),
	nl_stream(Out), 
	fail.
compile_all_predicates(Out) :-
	clause(compiled_resource(Instructions), _),
	write_asm(Out, Instructions),
	nl_stream(Out), 
	fail.
compile_all_predicates(Out) :-
	write_asm(Out, [fail: fail]),
	nl_stream(Out).

write_asm(_, []) :- !.
write_asm(Out, [Instruction|Instructions]) :- !,
	write_asm(Out, Instruction),
	write_asm(Out, Instructions).
write_asm(Out, comment(Comment)) :- !,
	write_stream(Out, '% '),
	writeq_stream(Out, Comment), nl_stream(Out).
write_asm(Out, debug(Comment)) :- !,
	tab_stream(Out, 8),
	write_stream(Out, '% '),
	writeq_stream(Out, Comment), nl_stream(Out).
write_asm(Out, (Label: Instruction)) :- !,
	writeq_stream(Out, Label), write_stream(Out, ' :'),
	nl_stream(Out),
	write_asm(Out, Instruction).
write_asm(Out, Instruction) :-
	tab_stream(Out, 8),
	writeq_stream(Out, Instruction), write_stream(Out, '.'),
	nl_stream(Out).

/****************************************************************
	Compile Linear Clauses
****************************************************************/
compile_linear_clauses([]) --> !,
	[init: proceed].
compile_linear_clauses(Clauses) -->
	[init: begin_imp(a(1))],
	compile_linear_clauses1(Clauses),
	[mid_imp(a(2),a(3))],
	[proceed].

compile_linear_clauses1([C]) --> !,
	compile_linear_clause(C).
compile_linear_clauses1([C|Cs]) -->
	compile_linear_clause(C),
	[more_imp],
	compile_linear_clauses1(Cs).

compile_linear_clause(C) -->
	{compile_clause((init :- C -<> true), Instructions0, [])},
	{Instructions0 = [_|Instructions1]},
	{llpc_append(Instructions, [_,_,_], Instructions1)},
	[Instructions].

/****************************************************************
	Compile Predicate
****************************************************************/
compile_predicate(Functor, Arity) -->
	{resource_predicates(Functor, Arity)},
	!,
	{functor(Head, Functor, Arity)},
	{findall((Head :- Body), internal_clause(Head, Body), Clauses)},
        compile_pred_R(Clauses, Functor/Arity).
compile_predicate(Functor, Arity) -->
	{functor(Head, Functor, Arity)},
	{findall((Head :- Body), internal_clause(Head, Body), Clauses)},
        compile_pred(Clauses, Functor/Arity).

compile_pred_R([], FA) --> 
	generate_res_switch1(FA), !.
compile_pred_R([Clause], FA) --> !,
	generate_res_switch2(FA),
	{copy_term(Clause, Clause1)},
	{numbervars(Clause1, 0, _)},
	[comment(Clause1)],
	[FA+prog: []],
	compile_clause(Clause).
compile_pred_R(Clauses, FA) --> 
	generate_res_switch2(FA),
	generate_switch_R(Clauses, FA),
        compile_pred2(Clauses, FA, 1).

compile_pred([], _) --> [], !.
compile_pred([Clause], FA) --> !,
	{copy_term(Clause, Clause1)},
	{numbervars(Clause1, 0, _)},
	[comment(Clause1)],
	[FA: []],
	compile_clause(Clause).
compile_pred(Clauses, FA) --> 
	generate_switch(Clauses, FA),
        compile_pred2(Clauses, FA, 1).

compile_pred2([], _, _) --> !.
compile_pred2([Clause|Clauses], FA, N) -->
	{copy_term(Clause, Clause1)},
	{numbervars(Clause1, 0, _)},
	[comment(Clause1)],
	[FA+N: []],
	compile_clause(Clause),
	{N1 is N + 1},
	compile_pred2(Clauses, FA, N1).

/*
	Generate pickup_resource, consume, execute_closure, 
	         try_resource, restore_resource, 
		 retry_resource_else, trust_resource, 
		 if_no_resource
*/
generate_res_switch1(FA) -->
	{FA = _F/A, I is A+1, J is A+2},
	[FA: []],
	[pickup_resource(FA, a(I), fail)],
	[if_no_resource(FA+res1)],
	[try_resource(FA+res1)],
	[FA+res0: []],
	[restore_resource],
	[pickup_resource(FA, a(I), FA+res2)],
	[if_no_resource(FA+res3)],
	[retry_resource_else(FA+res0)],
	[FA+res1: []],
	[consume(a(I), a(J))],
	[execute_closure(a(J))],
	[FA+res2: []],
	[trust_resource(fail)],
	[FA+res3: []],
	[trust_resource(FA+res1)].

generate_res_switch2(FA) -->
	{FA = _F/A, I is A+1, J is A+2},
	[FA: []],
	[pickup_resource(FA, a(I), FA+prog)],
	[try_resource(FA+res1)],
	[FA+res0: []],
	[restore_resource],
	[pickup_resource(FA, a(I), FA+res2)],
	[retry_resource_else(FA+res0)],
	[FA+res1: []],
	[consume(a(I), a(J))],
	[execute_closure(a(J))],
	[FA+res2: []],
	[trust_resource(FA+prog)].

/*
	Generate switch_on_term, try, retry and trust instructions
*/
generate_switch_R([], _) --> !.
generate_switch_R([_], FA) --> !,
	[FA+prog: []].
generate_switch_R(Clauses, F/0) --> !,
	[F/0+prog: []],
	{select_type(Clauses, var, 1, All)},
	generate_tries(All, F/0).
generate_switch_R(Clauses, FA) --> 
	[FA+prog: SWT],
	{select_type(Clauses, var, 1, All)},
	generate_tries(All, FA),
	{LV = FA+var},
	generate_sw(Clauses, int, FA, All, LI),
	generate_sw(Clauses, con, FA, All, LC),
	generate_sw(Clauses, str, FA, All, LS),
	generate_sw(Clauses, lis, FA, All, LL),
	{generate_sw_code(LV, LI, LC, LS, LL, SWT)}.

generate_switch([], _) --> !.
generate_switch([_], FA) --> !,
	[FA: []].
generate_switch(Clauses, F/0) --> !,
	[F/0: []],
	{select_type(Clauses, var, 1, All)},
	generate_tries(All, F/0).
generate_switch(Clauses, FA) --> 
	[FA: SWT],
	{select_type(Clauses, var, 1, All)},
	generate_tries(All, FA),
	{LV = FA+var},
	generate_sw(Clauses, int, FA, All, LI),
	generate_sw(Clauses, con, FA, All, LC),
	generate_sw(Clauses, str, FA, All, LS),
	generate_sw(Clauses, lis, FA, All, LL),
	{generate_sw_code(LV, LI, LC, LS, LL, SWT)}.

generate_sw_code(L, L, L, L, L, []) :- !.
generate_sw_code(LV, LI, LC, LS, LL, SWT) :-
	SWT = [switch_on_term(LV,LI,LC,LS,LL), LV: []].

generate_sw(Clauses, Type, FA, All, L) -->
	{select_type(Clauses, Type, 1, Ns)},
	{Ns = All},
	!,
	{L = FA+var}.
generate_sw(Clauses, Type, FA, _, L) -->
	{select_type(Clauses, Type, 1, Ns)},
	generate_sw1(Ns, Type, FA, L).

generate_sw1([], _, _, fail) --> !.
generate_sw1([N], _, FA, FA+N) --> !.
generate_sw1(Ns, Type, FA, L) -->
	{L = FA+Type},
	[L: []],
	generate_tries(Ns, FA).

generate_tries([N|Ns], FA) -->
	[try(FA+N)],
	generate_tries1(Ns, FA).

generate_tries1([N], FA) --> !,
	[trust(FA+N)].
generate_tries1([N|Ns], FA) --> !,
	[retry(FA+N)],
	generate_tries1(Ns, FA).

select_type([], _, _, []) :- !.
select_type([Clause|Clauses], Type, N, [N|Ns]) :-
	match_type(Clause, Type),
	!,
	N1 is N + 1,
	select_type(Clauses, Type, N1, Ns).
select_type([_|Clauses], Type, N, Ns) :-
	N1 is N + 1,
	select_type(Clauses, Type, N1, Ns).

match_type(_, var) :- !.
match_type((Head :- _), Type) :-
	arg(1, Head, A1),
	match_type1(A1, Type).

match_type1(A1, _) :- var(A1), !.
match_type1(A1, Type) :- integer(A1), !, Type=int.
match_type1(A1, Type) :- atom(A1), !, Type=con.
match_type1([_|_], Type) :- !, Type=lis.
match_type1(A1, Type) :- functor(A1, _, _), !, Type=str.

/****************************************************************
	Compile clause
****************************************************************/
compile_clause((_Head :- Body)) -->
	{nonvar(Body)}, {Body = '$ASM'(LLPAM)}, !,
	[LLPAM].
compile_clause((Head :- Body)) -->
	{precompile(Head, Body, Code)},
	{divide_into_chunks(Code, Chunks)},
	compile_chunks(Chunks),
	!.
compile_clause(Clause) -->
	{llpc_error([compilation,of,Clause,failed])},
	{fail}.

/****************************************************************
	Pre-compile
	get, put, are not compiled.
****************************************************************/
precompile(Head, Body, Code) :-
	precompile_head(Head, Code0, Code1),
	precompile_body(Body, Code1, []),
	flatten_code(Code0, Code, []).

precompile_head(Head) -->
	{Head =.. [_|Args]},
	precompile_head1(Args, 1).

precompile_head1([], _) --> !.
precompile_head1([A|As], I) -->
	[get(A, a(I))],
	{I1 is I + 1},
	precompile_head1(As, I1).

precompile_body(G, Code, Code1) :-
	precomp_body(G, Cut, Code0, Code1),
	precompile_body1(Cut, Code0, Code).

precompile_body1(Cut, Code, Code) :- var(Cut), !.
precompile_body1(cut(Level), Code, [get_level(Level)|Code]).

precomp_body(G, Cut) --> {var(G)}, !,
	precomp_body(call(G), Cut).
precomp_body(!, Cut) --> !,
	{Cut = cut(Level)},
	[cut(Level)].
precomp_body(true, _) --> !.
precomp_body(otherwise, _) --> !.
precomp_body(1, _) --> !.
precomp_body(fail, _) --> !, [fail].
precomp_body(false, _) --> !, [fail].
precomp_body(top, _) --> !, [top].
precomp_body(erase, _) --> !, [erase].
precomp_body((G1,G2), Cut) --> !,
	precomp_body(G1, Cut),
	precomp_body(G2, Cut).
precomp_body((G1&G2), Cut) --> !,
	[begin_with(L1)],
	precomp_body(G1, Cut),
	[mid_with(L2)],
	precomp_body(G2, Cut),
	[end_with(L1,L2)].
precomp_body((R-<>G), Cut) --> !,
	[begin_imp(L1)],
	precomp_res_R(R, lin),
	[mid_imp(L2,L3)],
	precomp_body(G, Cut),
	[end_imp(L1,L2,L3)].
precomp_body((R=>G), Cut) --> !,
	[begin_exp_imp(L1)],
	precomp_res_R(R, exp),
	[mid_exp_imp(L2,L3)],
	precomp_body(G, Cut),
	[end_exp_imp(L1,L2,L3)].
precomp_body(!G, Cut) --> !,
	[begin_bang(L1)],
	precomp_body(G, Cut),
	[end_bang(L1)].
precomp_body(G, _) -->
	precompile_builtin(G),
	!.
precomp_body(G, _) -->
	{G =.. [F|Args]},
	{length(Args, Arity)},
	precomp_call(F/Arity, Args, 1).

precomp_call(FA, [], _) --> !,
	[call(FA)].
precomp_call(FA, [A|As], I) -->
	[put(A, a(I))],
	{I1 is I + 1},
	precomp_call(FA, As, I1).

precomp_res_R(R, Type) --> {var(R)}, !,
	{(Type == lin -> X = '$add_res_R'(R) ; X = '$add_exp_res_R'(R))},
	precomp_body(X, _).
precomp_res_R((R1,R2), Type) --> !,
	precomp_res_R(R1, Type),
	{(Type == lin -> Instr = more_imp ; Instr = [])},
	[Instr],
	precomp_res_R(R2, Type).
precomp_res_R(R, Type) --> 
	precomp_res_S(R, Type).

precomp_res_S(R, Type) --> {var(R)}, !,
	{(Type == lin -> X = '$add_res_S'(R) ; X = '$add_exp_res_S'(R))},
	precomp_body(X, _).
precomp_res_S((R1&R2), Type) --> !,
	precomp_res_S(R1, Type),
	precomp_res_S(R2, Type).
precomp_res_S((forall Xs\ R), Type) --> {var(R)}, !,
	{(  Type == lin -> 
	     X = '$add_res_S'((forall Xs\ R)) 
	     ; 
	     X = '$add_exp_res_S'((forall Xs\ R))
	 )},
	precomp_body(X,_).
precomp_res_S((forall Xs\ G -<> R), Type) --> {var(R)}, !,
	{(  Type == lin -> 
	     X = '$add_res_S'((forall Xs\ G -<> R)) 
	     ; 
	     X = '$add_exp_res_S'((forall Xs\ G -<> R))
	 )},
	precomp_body(X, _).
precomp_res_S((G -<> R), Type) --> {var(R)}, !,
	{(  Type == lin -> 
	     X = '$add_res_S'((G -<> R)) 
	     ; 
	     X = '$add_exp_res_S'((G -<> R))
	 )},
	precomp_body(X, _).
precomp_res_S((forall Xs\ G -<> H), Type) --> {resource_head(H)}, !,
	{free_variables(Xs, (G -<> H), Freevars)},
	{functor(H, F, A)},
	{closure_label(F/A, Label)},
	{assert_compiled_resource((H :- G), Freevars, Label)},
	precomp_put(H, U),
	[put_closure(Label, Freevars, V)],
	{(Type == lin -> Instr = add_res(U, V) ; Instr = add_exp_res(U, V))},
	[Instr].
precomp_res_S((forall Xs\ H), Type) --> {resource_head(H)}, !,
	{free_variables(Xs, H, Freevars)},
	{functor(H, F, A)},
	{closure_label(F/A, Label)},
	{assert_compiled_resource((H :- true), Freevars, Label)},
	precomp_put(H, U),
	[put_closure(Label, Freevars, V)],
	{(Type == lin -> Instr = add_res(U, V) ; Instr = add_exp_res(U, V))},
	[Instr].
precomp_res_S((G -<> H), Type) --> {resource_head(H)}, !,
	{variables((G -<> H), Freevars)},
	{functor(H, F, A)},
	{closure_label(F/A, Label)},
	{assert_compiled_resource((H :- G), Freevars, Label)},
	precomp_put(H, U),
	[put_closure(Label, Freevars, V)],
	{(Type == lin -> Instr = add_res(U, V) ; Instr = add_exp_res(U, V))},
	[Instr].
precomp_res_S(H, Type) --> {resource_head(H)}, !,
	{variables(H, Freevars)},
	{functor(H, F, A)},
	{closure_label(F/A, Label)},
	{assert_compiled_resource((H :- true), Freevars, Label)},
	precomp_put(H, U),
	[put_closure(Label, Freevars, V)],
	{(Type == lin -> Instr = add_res(U, V) ; Instr = add_exp_res(U, V))},
	[Instr].
precomp_res_S(R, _) --> 
	{llpc_error([compilation,of,resource,'(',R,')',failed])},
	{fail}.

assert_compiled_resource(Clause, Freevars, Label) :-
	compile_resource(Clause, Freevars, Label, Instrs, []),
	assert(compiled_resource(Instrs)).

free_variables(Forall, Clause, Freevars) :-
	variables(Forall, V1),
	variables(Clause, V2),
	diff_list(V2, V1, Freevars).

closure_label(FA, FA+L) :- 
	inc_res_counter(ID),
	list_to_string([clo, ID], L).

inc_res_counter(N) :- 
	retract(res_counter(N)),
	!,
	N1 is N+1,
	assert(res_counter(N1)).
inc_res_counter(0) :- 
	llpc_error([dynamic,predicate,res_counter/1,is,not,asserted,yet]),
	assert(res_counter(0)).

/****************************************************************
	Compile resource clause
****************************************************************/
compile_resource(Clause, Freevars, Label) --> 
	{copy_term(Clause, Clause1)},
	{numbervars(Clause1, 0, _)},
	[comment(Clause1)],
	[Label : []],
	compile_res_clause(Clause, Freevars).

compile_res_clause((_Head :- Body), _) -->
	{nonvar(Body)}, 
	{Body = '$ASM'(LLPAM)}, 
	 !,
	[LLPAM].
compile_res_clause((Head :- Body), Freevars) -->
	{precompile(Head, Body, Code)},
	{divide_into_chunks([get_closure(Freevars)|Code], Chunks)},
	compile_chunks(Chunks),
	!.
compile_res_clause(Clause, Freevars) -->
	{llpc_error([compilation,of,resource,Clause,with,Freevars,failed])},
	{fail}.

/*
	Pre-compile of built-in predicate
*/
precompile_builtin(var(X)) --> !, precomp_builtin(var, X).
precompile_builtin(nonvar(X)) --> !, precomp_builtin(nonvar, X).
precompile_builtin(atomic(X)) --> !, precomp_builtin(atomic, X).
precompile_builtin(integer(X)) --> !, precomp_builtin(integer, X).
precompile_builtin(atom(X)) --> !, precomp_builtin(atom, X).
% precompile_builtin(X = Y) --> !, precomp_builtin((=), X, Y).
precompile_builtin(X == Y) --> !, precomp_builtin((==), X, Y).
precompile_builtin(X \== Y) --> !, precomp_builtin((\==), X, Y).
precompile_builtin(X @< Y) --> !, precomp_builtin((@<), X, Y).
precompile_builtin(X @> Y) --> !, precomp_builtin((@>), X, Y).
precompile_builtin(X @=< Y) --> !, precomp_builtin((@=<), X, Y).
precompile_builtin(X @>= Y) --> !, precomp_builtin((@>=), X, Y).
precompile_builtin(X =:= Y) --> !, precomp_arith((=:=), X, Y).
precompile_builtin(X =\= Y) --> !, precomp_arith((=\=), X, Y).
precompile_builtin(X > Y) --> !, precomp_arith((>), X, Y).
precompile_builtin(X =< Y) --> !, precomp_arith((=<), X, Y).
precompile_builtin(X < Y) --> !, precomp_arith((<), X, Y).
precompile_builtin(X >= Y) --> !, precomp_arith((>=), X, Y).
precompile_builtin(X is Y) --> !, [is(X, Y)].

precomp_builtin(B, X) -->
	precomp_put(X, U),
	[builtin(B, U)].

precomp_builtin(B, X, Y) -->
	precomp_put(X, U),
	precomp_put(Y, V),
	[builtin(B, U, V)].

precomp_arith(B, X, Y) -->
	precomp_is(U, X),
	precomp_is(V, Y),
	[builtin(B, U, V)].

precomp_put(X, X) --> {var(X)}, !.
precomp_put(X, U) -->
	[put(X, U)].

precomp_is(X, X) --> {var(X)}, !.
precomp_is(X, Y) -->
	[is(X, Y)].

/*
	Divide into Chunks
*/
divide_into_chunks(List, Chunks) :-
	div_into_chunks(List, Chunks).

div_into_chunks([], []) :- !.
div_into_chunks(List, [Chunk|Chunks]) :-
	extract_chunk(Chunk, List, Rest),
	!,
	div_into_chunks(Rest, Chunks).
div_into_chunks(List, [List]).

extract_chunk([call(G)]) -->
	[call(G)].
extract_chunk([G|Gs]) -->
	[G],
	extract_chunk(Gs).

/*
	Compile Chunks
*/
compile_chunks(Chunks) -->
	{alloc_y_regs(Chunks, 0, YN, [], Alloc0)},
	{max_reg(Max)},
	{YN =< Max+1},
	!,
	{alloc_voids(Chunks, Alloc0, Alloc)},
	{comp_chunks(Chunks, Alloc, Codes)},
	compile_chunks1(Codes, YN).
compile_chunks(_) -->
	{llpc_error([too,many,'Y',registers])},
	{fail}.
	 
compile_chunks1([Code], 0) --> !,
	change_last_call(Code).
compile_chunks1(Codes, YN) -->
	{needs_alloc(Codes, YN)},
	!,
	[allocate(YN)],
	flatten_code(Codes),
	[deallocate],
	[proceed].
compile_chunks1(Codes, _) -->
	flatten_code(Codes),
	[proceed].

needs_alloc(_, YN) :- YN > 0, !.
needs_alloc(Codes, _) :-
	llpc_member(Code, Codes),
	llpc_member(call(_), Code),
	!.

change_last_call([]) --> !,
	[proceed].
change_last_call([call(FN)|Code]) --> !,
	{Code = []},
	[execute(FN)].
change_last_call([C|Code]) -->
	[C],
	change_last_call(Code).

comp_chunks([], _, []) :- !.
comp_chunks([Chunk|Chunks], Alloc, [Code|Codes]) :-
	comp_chunk(Chunk, Alloc, Code, []),
	comp_chunks(Chunks, Alloc, Codes).

%comp_chunk(Chunk, Alloc) -->
%	comp_chunk0(Chunk, Alloc).
comp_chunk(Chunk, Alloc) -->
	{comp_chunk0(Chunk, Alloc, Instrs, [])},
	remove_redundant(Instrs).

comp_chunk0(Chunk, Alloc) -->
	{free_x_reg(Chunk, 1, XN)},
	comp_chunk1(Chunk, [XN,Alloc], _).

comp_chunk1([], VarInfo, VarInfo) --> !.
comp_chunk1([(L:[])|Cs], VarInfo0, VarInfo) --> !,
	[L:[]],
	comp_chunk1(Cs, VarInfo0, VarInfo).
comp_chunk1([(L:C)|Cs], VarInfo0, VarInfo) --> !,
	[L:[]],
	comp_chunk1([C|Cs], VarInfo0, VarInfo).
comp_chunk1([C|Cs], VarInfo0, VarInfo) --> !,
	comp_instr(C, VarInfo0, VarInfo1),
	comp_chunk1(Cs, VarInfo1, VarInfo).

free_x_reg([], XN, XN).
free_x_reg([get(_,V)|Cs], XN0, XN) :- nonvar(V), V = a(N), !,
	max(N+1, XN0, XN1),
	free_x_reg(Cs, XN1, XN).
free_x_reg([put(_,V)|Cs], XN0, XN) :- nonvar(V), V = a(N), !,
	max(N+1, XN0, XN1),
	free_x_reg(Cs, XN1, XN).
free_x_reg([_|Cs], XN0, XN) :-
	free_x_reg(Cs, XN0, XN).

remove_redundant([]) --> !.
remove_redundant([Instr|Instrs]) --> {redundant_instr(Instr)}, !,
	remove_redundant(Instrs).
%remove_redundant([get_var(a(X),a(Y)), put_val(a(X),a(Y))|Instrs]) --> !,
%	remove_redundant(Instrs).
remove_redundant([Instr|Instrs]) -->
	[Instr],
	remove_redundant(Instrs).

redundant_instr(get_var(X,X)).
redundant_instr(get_val(X,X)).
redundant_instr(put_val(X,X)).

/*
	Allocate Y Registers
*/
alloc_y_regs(Chunks, YN0, YN, Alloc0, Alloc) :-
	alloc_y_regs(Chunks, [], YN0, YN, Alloc0, Alloc).

alloc_y_regs([], _, YN, YN, Alloc, Alloc) :- !.
alloc_y_regs([Chunk|Chunks], PrevVars0, YN0, YN, Alloc0, Alloc) :-
	variables(Chunk, Vars),
	llpc_append(Vars, PrevVars0, PrevVars),
	alloc_y_regs(Chunks, PrevVars, YN0, YN1, Alloc0, Alloc1),
	alloc_y_reg1(Vars, PrevVars0, YN1, YN, Alloc1, Alloc).

alloc_y_reg1([], _, YN, YN, Alloc, Alloc) :- !.
alloc_y_reg1([V|Vs], PrevVars, YN0, YN, Alloc0, Alloc) :-
	memq(V, PrevVars),
	\+ allocated(Alloc0, V, _),
	!,
	YN1 is YN0 + 1,
	Alloc1 = [[V,y(YN1),_Seen]|Alloc0],
	alloc_y_reg1(Vs, PrevVars, YN1, YN, Alloc1, Alloc).
alloc_y_reg1([_|Vs], PrevVars, YN0, YN, Alloc0, Alloc) :-
	alloc_y_reg1(Vs, PrevVars, YN0, YN, Alloc0, Alloc).

/*
	Allocate Void Variables
*/
alloc_voids(Chunks, Alloc0, Alloc) :-
	variables(Chunks, Vars),
	alloc_voids1(Vars, Chunks, Alloc0, Alloc).

alloc_voids1([], _, Alloc, Alloc).
alloc_voids1([V|Vars], Chunks, Alloc0, Alloc) :-
	count_variable(V, Chunks, 1),
	!,
	Alloc1 = [[V,void,_Seen]|Alloc0],
	alloc_voids1(Vars, Chunks, Alloc1, Alloc).
alloc_voids1([_|Vars], Chunks, Alloc0, Alloc) :-
	alloc_voids1(Vars, Chunks, Alloc0, Alloc).

/*
	Compile instructions
	comp_instr(+Instr, +VarInfo0, -VarInfo)
	Instr:	Intermediate instruction
	VarInfo: [XN, Alloc]
	XN:	a(XN),a(XN+1),... are available
	Alloc:	[[Var,Type,Seen],...]
*/
comp_instr(get(X,V), VarInfo0, VarInfo) --> !,
	gen_get(X, V, VarInfo0, VarInfo).
comp_instr(get_closure(Vs), VarInfo0, VarInfo) --> !,
	gen_get_closure(Vs, VarInfo0, VarInfo).
comp_instr(put(X,V), VarInfo0, VarInfo) --> !,
	gen_put(X, V, VarInfo0, VarInfo).
comp_instr(put_closure(L, X, V), VarInfo0, VarInfo) --> !,
	gen_put_closure(L, X, V, VarInfo0, VarInfo).
comp_instr(is(V,X), VarInfo0, VarInfo) --> !,
	gen_is(X, V, VarInfo0, VarInfo).
comp_instr(builtin(B,X), VarInfo0, VarInfo) --> !,
	assign_temp(X, R1, _, VarInfo0, VarInfo),
	[builtin(B,R1)].
comp_instr(builtin(B,X,Y), VarInfo0, VarInfo) --> !,
	assign_temp(X, R1, _, VarInfo0, VarInfo1),
	assign_temp(Y, R2, _, VarInfo1, VarInfo),
	[builtin(B,R1,R2)].
comp_instr(add_res(X,Y), VarInfo0, VarInfo) --> !,
	assign_temp(X, R1, _, VarInfo0, VarInfo1),
	assign_temp(Y, R2, _, VarInfo1, VarInfo),
	[add_res(R1,R2)].
comp_instr(add_exp_res(X,Y), VarInfo0, VarInfo) --> !,
	assign_temp(X, R1, _, VarInfo0, VarInfo1),
	assign_temp(Y, R2, _, VarInfo1, VarInfo),
	[add_exp_res(R1,R2)].
comp_instr(begin_imp(X), VarInfo0, VarInfo) --> !,
	{assign_reg(X, R, _, VarInfo0, VarInfo)},
	[begin_imp(R)].
comp_instr(begin_exp_imp(X), VarInfo0, VarInfo) --> !,
	{assign_reg(X, R, _, VarInfo0, VarInfo)},
	[begin_exp_imp(R)].
comp_instr(mid_imp(X,Y), VarInfo0, VarInfo) --> !,
	{assign_reg(X, R1, _, VarInfo0, VarInfo1)},
	{assign_reg(Y, R2, _, VarInfo1, VarInfo)},
	[mid_imp(R1,R2)].
comp_instr(mid_exp_imp(X,Y), VarInfo0, VarInfo) --> !,
	{assign_reg(X, R1, _, VarInfo0, VarInfo1)},
	{assign_reg(Y, R2, _, VarInfo1, VarInfo)},
	[mid_exp_imp(R1,R2)].
comp_instr(end_imp(X,Y,Z), VarInfo0, VarInfo) --> !,
	{assign_reg(X, R1, _, VarInfo0, VarInfo1)},
	{assign_reg(Y, R2, _, VarInfo1, VarInfo2)},
	{assign_reg(Z, R3, _, VarInfo2, VarInfo)},
	[end_imp(R1,R2,R3)].
comp_instr(end_exp_imp(X,Y,Z), VarInfo0, VarInfo) --> !,
	{assign_reg(X, R1, _, VarInfo0, VarInfo1)},
	{assign_reg(Y, R2, _, VarInfo1, VarInfo2)},
	{assign_reg(Z, R3, _, VarInfo2, VarInfo)},
	[end_exp_imp(R1,R2,R3)].
comp_instr(begin_with(X), VarInfo0, VarInfo) --> !,
	{assign_reg(X, R, _, VarInfo0, VarInfo)},
	[begin_with(R)].
comp_instr(mid_with(X), VarInfo0, VarInfo) --> !,
	{assign_reg(X, R, _, VarInfo0, VarInfo)},
	[mid_with(R)].
comp_instr(end_with(X,Y), VarInfo0, VarInfo) --> !,
	{assign_reg(X, R1, _, VarInfo0, VarInfo1)},
	{assign_reg(Y, R2, _, VarInfo1, VarInfo)},
	[end_with(R1,R2)].
comp_instr(begin_bang(X), VarInfo0, VarInfo) --> !,
	{assign_reg(X, R, _, VarInfo0, VarInfo)},
	[begin_bang(R)].
comp_instr(end_bang(X), VarInfo0, VarInfo) --> !,
	{assign_reg(X, R, _, VarInfo0, VarInfo)},
	[end_bang(R)].
comp_instr(get_level(X), VarInfo0, VarInfo) -->
	{assign_reg(X, y(N), _, VarInfo0, VarInfo)},
	!,
	[get_level(y(N))].
comp_instr(get_level(_), VarInfo, VarInfo) --> !.
comp_instr(cut(X), VarInfo0, VarInfo) -->
	{assign_reg(X, y(N), _, VarInfo0, VarInfo)},
	!,
	[cut(y(N))].
comp_instr(cut(_), VarInfo, VarInfo) --> !,
	[neck_cut].
comp_instr(Instr, VarInfo, VarInfo) -->
	[Instr].

/*
	Generate get/unify instructions
*/
gen_get(X, V, VI0, VI) -->
	gen_get([V=X], VI0, VI).

gen_get([], VI, VI) --> !.
gen_get([A=X|_], VI, VI) --> {var(A)}, !,
	{llpc_error([get(X,A),error])},
	{fail}.
gen_get([A=X|Instrs], VI0, VI) --> {var(X)}, !,
	{assign_reg(X, R, Seen, VI0, VI1)},
	gen_get_var(R, Seen, A),
	gen_get(Instrs, VI1, VI).
gen_get([_=X|_], _, _) --> {float(X)}, !,
	{llpc_error([floating,point,numbers,are,not,supported])}, 
	{fail}.
gen_get([A=X|Instrs], VI0, VI) --> {integer(X)}, !,
	[get_int(X, A)],
	gen_get(Instrs, VI0, VI).
gen_get([A=X|Instrs], VI0, VI) --> {atom(X)}, !,
	[get_con(X, A)],
	gen_get(Instrs, VI0, VI).
gen_get([A=X|Instrs], VI0, VI) --> {ground(X)}, {\+ atomic(X)}, !,
	[get_ground(X, A)],
	gen_get(Instrs, VI0, VI).
gen_get([A=X|Instrs], VI0, VI) --> {X = [X1|X2]}, !,
	[get_list(A)],
	gen_unify([X1,X2], Instrs1, VI0, VI1),
	gen_get(Instrs1, VI1, VI2),
	gen_get(Instrs, VI2, VI).
gen_get([A=X|Instrs], VI0, VI) -->
	{X =.. [F|Args]},
	{length(Args, Arity)},
	[get_str(F/Arity,A)],
	gen_unify(Args, Instrs1, VI0, VI1),
	gen_get(Instrs1, VI1, VI2),
	gen_get(Instrs, VI2, VI).

gen_get_var(void, _, _) --> !.
gen_get_var(R, Seen, A) --> {var(Seen)}, !,
	{Seen = yes},
	[get_var(R,A)].
gen_get_var(R, _, A) -->
	[get_val(R,A)].

gen_unify([], [], VI, VI) --> !.
gen_unify([X|Xs], Instrs, VI0, VI) --> {var(X)}, !,
	{assign_reg(X, R, Seen, VI0, VI1)},
	gen_unify_var(R, Seen),
	gen_unify(Xs, Instrs, VI1, VI).
gen_unify([X|_], _, _, _) --> {float(X)}, !,
	{llpc_error([floating,point,numbers,are,not,supported])},
	{fail}.
gen_unify([X|Xs], Instrs, VI0, VI) --> {integer(X)}, !,
	[unify_int(X)],
	gen_unify(Xs, Instrs, VI0, VI).
gen_unify([X|Xs], Instrs, VI0, VI) --> {atom(X)}, !,
	[unify_con(X)],
	gen_unify(Xs, Instrs, VI0, VI).
gen_unify([X|Xs], Instrs, VI0, VI) --> {ground(X)}, {\+ atomic(X)}, !,
	[unify_ground(X)],
	gen_unify(Xs, Instrs, VI0, VI).
gen_unify([X|Xs], [R=X|Instrs], VI0, VI) -->
	{assign_reg(X, R, Seen, VI0, VI1)},
	gen_unify_var(R, Seen),
	gen_unify(Xs, Instrs, VI1, VI).

gen_unify_var(void, _) --> !,
	[unify_void(1)].
gen_unify_var(R, Seen) --> {var(Seen)}, !,
	{Seen = yes},
	[unify_var(R)].
gen_unify_var(R, _) -->
	[unify_val(R)].

/*
	Generate unify_val & put_closure L, N, Ai
*/
gen_get_closure(Vs, VI0, VI) -->
	gen_unify(Vs, [], VI0, VI).

gen_put_closure(Label, Freevars, X, VI0, VI) --> {var(X)}, !,
	{length(Freevars, N)},
	{assign_reg(X, R, Seen, VI0, VI1)},
	{Seen = yes},
	gen_put_args(Freevars, Vs, VI1, VI2),
	[put_closure(Label, N, R)],
	gen_set(Vs, VI2, VI).

/*
	Generate put/set instructions
*/
gen_put(X, A, VI0, VI) --> {var(A)}, !,
	{assign_reg(A, R, Seen, VI0, VI1)},
	gen_put(X, R, VI1, VI),
	{Seen = yes}.
gen_put(X, A, VI0, VI) --> {var(X)}, !,
	{assign_reg(X, R, Seen, VI0, VI)},
	gen_put_var(R, Seen, A).
gen_put(X, _, _, _) --> {float(X)}, !,
	{llpc_error([floating,point,numbers,are,not,supported])},
	{fail}.
gen_put(X, A, VI, VI) --> {integer(X)}, !,
	[put_int(X, A)].
gen_put(X, A, VI, VI) --> {atom(X)}, !,
	[put_con(X, A)].
gen_put(X, A, VI, VI) --> {ground(X)}, {\+ atomic(X)}, !,
	[put_ground(X, A)].
gen_put(X, A, VI0, VI) --> {X = [X1|X2]}, !,
	gen_put_args([X1,X2], Vs, VI0, VI1),
	[put_list(A)],
	gen_set(Vs, VI1, VI).
gen_put(X, A, VI0, VI) -->
	{X =.. [F|Args]},
	{length(Args, Arity)},
	gen_put_args(Args, Vs, VI0, VI1),
	[put_str(F/Arity,A)],
	gen_set(Vs, VI1, VI).

gen_put_args([], [], VI, VI) --> !.
gen_put_args([X|Xs], [X|Vs], VI0, VI) --> {var(X)}, !,
	gen_put_args(Xs, Vs, VI0, VI).
gen_put_args([X|Xs], [X|Vs], VI0, VI) --> {atomic(X)}, !,
	gen_put_args(Xs, Vs, VI0, VI).
gen_put_args([X|Xs], [V|Vs], VI0, VI) -->
	{assign_reg(V, R, Seen, VI0, VI1)},
	gen_put(X, R, VI1, VI2),
	{Seen = yes},
	gen_put_args(Xs, Vs, VI2, VI).

gen_put_var(void, _, A) --> !,
	[put_var(A,A)].
gen_put_var(R, Seen, A) --> {var(Seen)}, !,
	{Seen = yes},
	[put_var(R,A)].
gen_put_var(R, _, A) -->
	[put_val(R,A)].

gen_set([], VI, VI) --> !.
gen_set([X|Xs], VI0, VI) --> {var(X)}, !,
	{assign_reg(X, R, Seen, VI0, VI1)},
	gen_set_var(R, Seen),
	gen_set(Xs, VI1, VI).
gen_set([X|_], _, _) --> {float(X)}, !,
	{llpc_error([floating,point,numbers,are,not,supported])},
	{fail}.
gen_set([X|Xs], VI0, VI) --> {integer(X)}, !,
	[set_int(X)],
	gen_set(Xs, VI0, VI).
gen_set([X|Xs], VI0, VI) --> {atom(X)}, !,
	[set_con(X)],
	gen_set(Xs, VI0, VI).
gen_set([X|Xs], VI0, VI) --> {ground(X)}, {\+ atomic(X)}, !,
	[set_ground(X)],
	gen_set(Xs, VI0, VI).

gen_set_var(void, _) --> !,
	[set_void(1)].
gen_set_var(R, Seen) --> {var(Seen)}, !,
	{Seen = yes},
	[set_var(R)].
gen_set_var(R, _) -->
	[set_val(R)].

/*
	Generate arithmetic instructions
*/
gen_is(X, A, VI0, VI) --> {var(X)}, !,
	gen_arith(is, X, A, VI0, VI).
gen_is(X, _, _, _) --> {float(X)}, !,
	{llpc_error([floating,point,numbers,are,not,supported])},
	{fail}.
gen_is(X, A, VI0, VI) --> {integer(X)}, !,
	gen_put(X, A, VI0, VI).
gen_is([X], A, VI0, VI) --> !,
	gen_is1(X, A, VI0, VI).
gen_is(+(X), A, VI0, VI) --> !,
	gen_is1(X, A, VI0, VI).
gen_is(X+Y, A, VI0, VI) --> !, 
	gen_is1(X, U, VI0, VI1),
	gen_is1(Y, V, VI1, VI2), 
	gen_arith((+), U, V, A, VI2, VI).
gen_is(-X, A, VI0, VI) --> !,
	gen_is1(X, U, VI0, VI1), 
	gen_arith((-), U, A, VI1, VI).
gen_is(X-Y, A, VI0, VI) --> !,
	gen_is1(X, U, VI0, VI1),
	gen_is1(Y, V, VI1, VI2), 
	gen_arith((-), U, V, A, VI2, VI).
gen_is(X*Y, A, VI0, VI) --> !,
	gen_is1(X, U, VI0, VI1),
	gen_is1(Y, V, VI1, VI2), 
	gen_arith((*), U, V, A, VI2, VI).
gen_is(X/Y, A, VI0, VI) --> !,
	gen_is1(X, U, VI0, VI1),
	gen_is1(Y, V, VI1, VI2), 
	gen_arith((/), U, V, A, VI2, VI).
gen_is(X//Y, A, VI0, VI) --> !,
	gen_is1(X, U, VI0, VI1),
	gen_is1(Y, V, VI1, VI2), 
	gen_arith((//), U, V, A, VI2, VI).
gen_is(X mod Y, A, VI0, VI) --> !,
	gen_is1(X, U, VI0, VI1),
	gen_is1(Y, V, VI1, VI2), 
	gen_arith(mod, U, V, A, VI2, VI).
gen_is(X<<Y, A, VI0, VI) --> !,
	gen_is1(X, U, VI0, VI1),
	gen_is1(Y, V, VI1, VI2), 
	gen_arith((<<), U, V, A, VI2, VI).
gen_is(X>>Y, A, VI0, VI) --> !,
	gen_is1(X, U, VI0, VI1),
	gen_is1(Y, V, VI1, VI2), 
	gen_arith((>>), U, V, A, VI2, VI).
gen_is(\(X), A, VI0, VI) --> !,
	gen_is1(X, U, VI0, VI1), 
	gen_arith((\), U, A, VI1, VI).
gen_is(X/\Y, A, VI0, VI) --> !,
	gen_is1(X, U, VI0, VI1),
	gen_is1(Y, V, VI1, VI2), 
	gen_arith((/\), U, V, A, VI2, VI).
gen_is(X\/Y, A, VI0, VI) --> !,
	gen_is1(X, U, VI0, VI1),
	gen_is1(Y, V, VI1, VI2), 
	gen_arith((\/), U, V, A, VI2, VI).
gen_is(X#Y, A, VI0, VI) --> !,
	gen_is1(X, U, VI0, VI1),
	gen_is1(Y, V, VI1, VI2), 
	gen_arith((#), U, V, A, VI2, VI).
gen_is(abs(X), A, VI0, VI) --> !,
	gen_is1(X, U, VI0, VI1), 
	gen_arith(abs, U, A, VI1, VI).
gen_is(max(X,Y), A, VI0, VI) --> !,
	gen_is1(X, U, VI0, VI1),
	gen_is1(Y, V, VI1, VI2), 
	gen_arith(max, U, V, A, VI2, VI).
gen_is(min(X,Y), A, VI0, VI) --> !,
	gen_is1(X, U, VI0, VI1),
	gen_is1(Y, V, VI1, VI2), 
	gen_arith(min, U, V, A, VI2, VI).
gen_is(X, _, _, _) -->
	{llpc_error([unknown,arithemetic,expression,X])},
	{fail}.

gen_is1(X, A, VI, VI) --> {var(X)}, {var(A)}, !,
	{X = A}.
gen_is1(X, A, VI0, VI) -->
	gen_is(X, A, VI0, VI).

gen_arith(FA, X, Y, VI0, VI) -->
	assign_temp(X, R1, _, VI0, VI1),
	assign_temp(Y, R2, Seen, VI1, VI),
	[builtin(FA, R1, R2)],
	{Seen = yes}.

gen_arith(FA, X, Y, Z, VI0, VI) -->
	assign_temp(X, R1, _, VI0, VI1),
	assign_temp(Y, R2, _, VI1, VI2),
	assign_temp(Z, R3, Seen, VI2, VI),
	[builtin(FA, R1, R2, R3)],
	{Seen = yes}.

/****************************************************************
	Utilities
****************************************************************/
read_stream(In, X) :-
	current_input(Stream),
	set_input(In),
	read(X),
	set_input(Stream).

write_stream(Out, X) :-
	current_output(Stream),
	set_output(Out),
	write(X),
	set_output(Stream).

writeq_stream(Out, X) :-
	current_output(Stream),
	set_output(Out),
	writeq(X),
	set_output(Stream).

writec_stream(Out, X) :-
	current_output(Stream),
	set_output(Out),
	write_canonical(X),
	set_output(Stream).

tab_stream(Out, X) :-
	current_output(Stream),
	set_output(Out),
	tab(X),
	set_output(Stream).

nl_stream(Out) :-
	current_output(Stream),
	set_output(Out),
	nl,
	set_output(Stream).

llpc_error(M) :-
	llpc_message(['***','LLPC','Error'|M]).

llpc_message([]) :- nl, flush_output(user_output).
llpc_message([M|Ms]) :-
	write(M), write(' '),
	llpc_message(Ms).

max_reg(1023).

assign_temp(X, R, Seen, VI0, VI) --> {nonvar(X)}, !,
	{assign_reg(_, R, Seen, VI0, VI)},
	gen_put(X, R, VI0, VI).
assign_temp(X, R, Seen, VI0, VI) -->
	{assign_reg(X, R0, Seen0, VI0, VI1)},
	assign_temp1(R0, Seen0, R, Seen, VI1, VI),
	{Seen = yes}.

assign_temp1(void, _, R, Seen, VI0, VI) --> !,
	{assign_reg(_, R, Seen, VI0, VI)},
	[put_var(R,R)].
assign_temp1(y(N), Seen0, R, Seen, VI0, VI) --> {var(Seen0)}, !,
	{Seen0 = yes},
	{assign_reg(_, R, Seen, VI0, VI)},
	[put_var(y(N),R)].
assign_temp1(y(N), _, R, Seen, VI0, VI) --> !,
	{assign_reg(_, R, Seen, VI0, VI)},
	[put_val(y(N),R)].
assign_temp1(R, Seen, R, Seen, VI, VI) --> {var(Seen)}, !,
	[put_var(R,R)].
assign_temp1(R, Seen, R, Seen, VI, VI) --> !.

assign_reg(X, R, Seen, [XN0,Alloc0], [XN0,Alloc0]) :-
	allocated(Alloc0, X, [R,Seen]),
	!.
assign_reg(X, R, Seen, [XN0,Alloc0], [XN,Alloc]) :-
	max_reg(Max),
	XN0 =< Max,
	!,
	R = a(XN0),
	XN is XN0 + 1,
	Alloc = [[X,R,Seen]|Alloc0].
assign_reg(_, _, _, _, _) :-
	llpc_error([too,many,'A',registers]),
	fail.

allocated([[V|X]|_], V0, X) :-
	V == V0,
	!.
allocated([_|Alloc], V0, X) :-
	allocated(Alloc, V0, X).

flatten_code([]) --> !.
flatten_code([(L: C)|Code]) --> !,
	[L: []],
	flatten_code([C|Code]).
flatten_code([Code1|Code2]) --> !,
	flatten_code(Code1),
	flatten_code(Code2).
flatten_code(Code) -->
	[Code].

llpc_reverse(Xs, Zs) :- llpc_rev(Xs, [], Zs).

llpc_rev([], Zs, Zs).
llpc_rev([X|Xs], Ys, Zs) :- llpc_rev(Xs, [X|Ys], Zs).

llpc_append([], Zs, Zs).
llpc_append([X|Xs], Ys, [X|Zs]) :- llpc_append(Xs, Ys, Zs).

variables(X, Vs) :- variables(X, [], Vs).

variables(X, Vs, Vs) :- var(X), memq(X, Vs), !.
variables(X, Vs, [X|Vs]) :- var(X), !.
variables(X, Vs0, Vs0) :- atomic(X), !.
variables([X|Xs], Vs0, Vs) :- !,
	variables(X, Vs0, Vs1), variables(Xs, Vs1, Vs).
variables(X, Vs0, Vs) :- X =.. Xs, variables(Xs, Vs0, Vs).

count_variable(V, X, 1) :- V == X, !.
count_variable(_, X, 0) :- var(X), !.
count_variable(_, X, 0) :- atomic(X), !.
count_variable(V, [X|Y], N) :- !,
	count_variable(V, X, N1),
	count_variable(V, Y, N2),
	N is N1 + N2.
count_variable(V, X, N) :-
	X =.. Xs,
	count_variable(V, Xs, N).

diff_list([], _, []).
diff_list([X|Xs], Ys, Zs) :- memq(X, Ys), !, diff_list(Xs, Ys, Zs).
diff_list([X|Xs], Ys, [X|Zs]) :- diff_list(Xs, Ys, Zs).

memq(X, [Y|_]) :- X==Y, !.
memq(X, [_|Ys]) :- memq(X, Ys).

llpc_member(X, [X|_]).
llpc_member(X, [_|Ys]) :- llpc_member(X, Ys).

dlist([]) --> [].
dlist([X|Y]) --> [X], dlist(Y).

max(X, Y, Z) :- X1 is X, Y1 is Y, max1(X1, Y1, Z).

max1(X, Y, X) :- X >= Y, !.
max1(_, Y, Y).

list_to_string([], S) :- !, S = ''.
list_to_string([X|Xs], S) :- 
	name(X, L1),
	list_to_string(Xs, S1),
	name(S1, L2),
	llpc_append(L1, L2, L3),
	name(S, L3).

/*********************************************************************
  expand_resource(+Clause, -NewClause, -Functor)
	+Clause      ---   a clause of LLP program
        -NewClause   ---   new clause of LLP program
        -Functor     ---   list of resource functors(f/n) in Clause.
*********************************************************************/
/* 
  Preprocessing phase to expand resource clause such that:
    * Goal 
	R1 -<> (R2 -<> G)   --> (R1,R2) -<> G
	R1 => (R2 => G)     --> (R1 & R2) => G
    * Resource
        R :- G              --> G -<> R
        R <= G              --> G => R
        G => R              --> (!G) -<> R
        G -<> (R1 & R2)     --> (G -<> R1) & (G -<> R2)
        G1 -<> (G2 -<> R)   --> (G1, G2) -<> R
        G -<> (forall X\R)  --> forall Z\ G -<> R[Z/X]
        (Z is a new variable that doesn't occur in G)
        forall X\ R1 & R2   --> (forall X\R1) & (forall X\R2)
        (R1 & R2) & R3      --> R1 & (R2 & R3)
*/
expand_resource(Clause, NewClause) :-
	expand_resource(Clause, NewClause, _).

expand_resource(Clause, NewClause, Functor) :-
	expand_res(Clause, Clause1, Functor1),
	NewClause = Clause1,
	Functor = Functor1.

expand_res(Cl, Cl, []) :- var(Cl), !.
expand_res(Cl, Cl1, F) :-  
	llp_expansion(Cl, Cl1, F), !.
expand_res(Cl, Cl, []).

llp_expansion((H :- B0), (H :- B), FA) :- !,
	body_expansion(B0, B, FA0),
	flatten_list(FA0, FA, []).
llp_expansion(H, H, []).

body_expansion(G, G, []) :- var(G), !.
body_expansion((G1, G2), (GG1, GG2), [FA1|FA2]) :- !,
	body_expansion(G1, GG1, FA1),
	body_expansion(G2, GG2, FA2).
body_expansion((G1 & G2), (GG1 & GG2), [FA1|FA2]) :- !,
	body_expansion(G1, GG1, FA1),
	body_expansion(G2, GG2, FA2).
body_expansion((G1 ; G2), (GG1 ; GG2), [FA1|FA2]) :- !,
	body_expansion(G1, GG1, FA1),
	body_expansion(G2, GG2, FA2).
%%%
body_expansion((R -<> G), GG, FA) :- nonvar(R), R = (!R1),!,
	body_expansion((R1 => G), GG, FA).
%%%
body_expansion((R -<> G), (RR -<> GG), [FA1|FA2]) :- !,
	limp_expansion(G, R, G1, RR, FA1),
	body_expansion(G1, GG, FA2).
body_expansion((R => G), (RR => GG), [FA1|FA2]) :- !,
	iimp_expansion(G, R, G1, RR, FA1),
	body_expansion(G1, GG, FA2).
body_expansion((!G), (!GG), F) :- !,
	body_expansion(G, GG, F).
body_expansion(G, _, _) :- G = (forall _\ _), !,
	llpc_error([goal,G,is,invalid]),
	fail.
body_expansion(G, G, []).

limp_expansion(G, R, G, RR, F) :- var(G), !,
	translate_R(R, RR, F).
limp_expansion((R0 -<> G0), R, (R1 => G0), RR, F) :- nonvar(R0), R0 = (!R1), !,
	translate_R(R, RR, F).
limp_expansion((R0 -<> G0), R, GG, RR, F) :- !,
	limp_expansion(G0, (R,R0), GG, RR, F).
limp_expansion(G, R, G, RR, F) :- 
	translate_R(R, RR, F).

iimp_expansion(G, R, G, RR, F) :- var(G), !,
	translate_R(R, RR, F).
%%%
iimp_expansion((R0 -<> G0), R, GG, RR, F) :- nonvar(R0), R0 = (!R1), !,
	iimp_expansion(G0, (R & R1), GG, RR, F).
%%%
iimp_expansion((R0 => G0), R, GG, RR, F) :- !,
	iimp_expansion(G0, (R & R0), GG, RR, F).
iimp_expansion(G, R, G, RR, F) :- 
	translate_R(R, RR, F).

translate_R(R, R, []) :- var(R), !.
translate_R((R1, R2), RR, F) :- 
	nonvar(R1), 
	R1 = (R3, R4),
	!,
	translate_R((R3, (R4, R2)), RR, F).
translate_R((R1, R2), (RR1, RR2), [FA1|FA2]) :- !,
	translate_R(R1, RR1, FA1),
	translate_R(R2, RR2, FA2).
translate_R(R, RR, F) :-
	translate_S(R, R1, F),
	flatten_S(R1, RR).

translate_S(R, R, []) :- var(R), !.
translate_S((G -<> R), (G -<> R), []) :- var(R), !.


translate_S((forall Xs\ R), RR, []) :- var(R), !,
	variable_as_head(R, Xs, (forall Xs\ R)),
	rm_redundant_forallvars((forall Xs\ R), RR).
translate_S((forall Xs\ G -<> R), RR, []) :- var(R), !,
	variable_as_head(R, Xs, (forall Xs\ G -<> R)),
	rm_redundant_forallvars((forall Xs\ G -<> R), RR).


translate_S((R :- G), RR, F) :- !,
	translate_S((G -<> R), RR, F).
translate_S((G => R), RR, F) :- !,
	translate_S(((!G) -<> R), RR, F).
translate_S((R <= G), RR, F) :- !,
	translate_S(((!G) -<> R), RR, F).
translate_S((R1 & R2), (RR1 & RR2), [FA1|FA2]) :- !,
	translate_S(R1, RR1, FA1),
	translate_S(R2, RR2, FA2).
translate_S((G -<> (R1 & R2)), RR, F) :- !,
	translate_S(((G -<> R1) & (G -<> R2)), RR, F).
translate_S((G1 -<> (G2 -<> R)), RR, F) :- !,
	translate_S(((G1, G2) -<> R), RR, F).
translate_S((G1 -<> (R :- G2)), RR, F) :- !,
	translate_S(((G1, G2) -<> R), RR, F).
translate_S((G1 -<> (G2 => R)), RR, F) :- !,
	translate_S(((G1, (!G2)) -<> R), RR, F).
translate_S((G1 -<> (R <= G2)), RR, F) :- !,
	translate_S(((G1, (!G2)) -<> R), RR, F). 
translate_S((G -<> (forall Xs\ R)), RR, F) :- !,
	replace_forallvars(Xs, R, Ys, R1),
	translate_S((forall Ys\ G -<> R1), RR, F).
translate_S((forall Xs\ R :- G), RR, F) :- !,
	translate_S((forall Xs\ G -<> R), RR, F).
translate_S((forall Xs\ G => R), RR, F) :- !,
	translate_S((forall Xs\ (!G) -<> R), RR, F).
translate_S((forall Xs\ R <= G), RR, F) :- !,
	translate_S((forall Xs\ (!G) -<> R), RR, F).
translate_S((forall Xs\ R1 & R2), RR, F) :- !,
	translate_S(((forall Xs\R1) & (forall Xs\R2)), RR, F).
translate_S((forall Xs\ G -<> (R1 & R2)), RR, F) :- !,
	translate_S(((forall Xs\ G -<> R1) & (forall Xs\ G -<> R2)), RR, F).
translate_S((forall Xs\ G1 -<> (G2 -<> R)), RR, F) :- !,
	translate_S((forall Xs\ (G1, G2) -<> R), RR, F).
translate_S((forall Xs\ G1 -<> (R :- G2)), RR, F) :- !,
	translate_S((forall Xs\ (G1, G2) -<> R), RR, F).
translate_S((forall Xs\ G1 -<> (G2 => R)), RR, F) :- !,
	translate_S((forall Xs\ (G1, (!G2)) -<> R), RR, F).
translate_S((forall Xs\ G1 -<> (R <= G2)), RR, F) :- !,
	translate_S((forall Xs\ (G1, (!G2)) -<> R), RR, F).
translate_S((forall Xs\ G -<> (forall Ys\ R)), RR, F) :- !,
	replace_forallvars(Ys, R, Zs, R1),
	translate_S((forall Xs,Zs\G -<> R1), RR, F).
translate_S((forall Xs\ (forall Ys\ R)), RR, F) :- !,
	translate_S((forall Xs,Ys\ R), RR, F).
translate_S((forall Xs\ G -<> H), RR, [F/A]) :- resource_head(H), !,
	replace_forallvars(Xs, (G -<> H), Ys, R1),
	rm_redundant_forallvars((forall Ys\ R1), RR),
	functor(H, F, A).
translate_S((forall Xs\ H), RR, [F/A]) :- resource_head(H), !,
	replace_forallvars(Xs, H, Ys, R1),
	rm_redundant_forallvars((forall Ys\ R1), RR),
	functor(H, F, A).
translate_S((G -<> H), (G -<> H), [F/A]) :- resource_head(H), !,
	functor(H, F, A).
translate_S(H, H, [F/A]) :- resource_head(H), !,
	functor(H, F, A).
translate_S(R, _, _) :-
	llpc_error([resource,'(',R,')',is,invalid]),
	fail.

flatten_S(R, R) :- var(R), !.
flatten_S((R1 & R2), RR) :- 
	nonvar(R1), 
	R1 = (R3 & R4),
	!,
	flatten_S((R3 & (R4 & R2)), RR).
flatten_S((R1 & R2), (RR1 & RR2)) :- !,
	flatten_S(R1, RR1),
	flatten_S(R2, RR2).
flatten_S(R, R).

resource_head(H) :-
	\+ (H = (_ & _)),
	\+ (H = (_ -<> _)),
	\+ (H = (_ :- _)),
	\+ (H = (_ => _)),
	\+ (H = (_ <= _)),
	\+ (H = (forall _ \ _)),
	\+ (H = !(_)),
	\+ (H = (_ , _)),
	\+ (H = (_ ; _)).

variable_as_head(H, Xs, R) :- 
	variables(Xs,Ys), 
	memq(H, Ys), 
	!,
	llpc_error([head,H,can,not,be,quantified,in,R]),
	fail.
variable_as_head(_, _, _).

replace_forallvars(Xs, R, Ys, RR) :-
	variables(Xs, L1),
	replace_variable(L1, L2, R, RR),
	list_to_conj(L2, Ys).

replace_variable([], [], R, R) :- !.
replace_variable([X|Xs], [Y|Ys], R, RR) :-
	replace_var(X, Y, R, R1),
	replace_variable(Xs, Ys, R1, RR).

replace_var(X, Y, R, RR) :- var(R), R == X, !, RR = Y.
replace_var(_, _, R, R) :- var(R), !.
replace_var(_, _, R, R) :- atomic(R), !.
replace_var(X, Y, [R|Rs], [RR|RRs]) :- !,
	replace_var(X, Y, R, RR),
	replace_var(X, Y, Rs, RRs).
replace_var(X, Y, R, RR) :- !,	
	R =.. L1,
	replace_var(X, Y, L1, L2),
	RR =.. L2.

rm_redundant_forallvars((forall X\ Y), R) :-
	variables(X,Xs),
	variables(Y,Ys),
	diff_list(Ys, Xs, Zs),
	diff_list(Ys, Zs, Ws),
	( Ws == [] -> R = (Y) ; 
	    list_to_conj(Ws, X1), R = (forall X1\ Y)
	).

list_to_conj(X, _) :-  var(X), !,
	llpc_error([variable,X,can,not,be,converted,to,'(A,B)',expression]),
	fail.
list_to_conj([], _) :- !,
	llpc_error(['[]',can,not,be,converted,to,'(A,B)',expression]),
	fail.
list_to_conj([X], (X)) :- !.
list_to_conj([X,Y], (X,Y)) :- !.
list_to_conj([X|Ys], (X,Zs)) :- 
	list_to_conj(Ys, Zs).

conj_to_list(X, _) :-  var(X), !,
	llpc_error([variable,X,can,not,be,converted,to,'(A,B)',expression]),
	fail.
conj_to_list(((X1,X2),Xs), Y) :- !,
	conj_to_list((X1,(X2,Xs)), Y).
conj_to_list((X,Xs), [X|Zs]) :- !,
	conj_to_list(Xs, Zs).
conj_to_list((X), [X]).

flatten_list([]) --> !.
flatten_list([L1|L2]) --> !,
	flatten_list(L1),
	flatten_list(L2).
flatten_list(L) --> [L].

/*********************************************************************
  eliminate_disj(Cl, NewCl, DummyCls)
        Cl          ---   a clause
        NewCl       ---   a new clause
        DummyCls    ---   list of dummy clauses
*********************************************************************/
/*
  For example,
  | ?- eliminate_disj((a(X) :- b(X);c(X)), NewCl, DummyCls).
  NewCl = a(X) :- dummy(X),
  DummyCls = [(dummy(A) :- b(A)),(dummy(B) :- c(B))] ? 

  Preprocessing phase to eliminate disjunctions from the clause that
  contains (A;B)  or  (IF->THEN;ELSE)  or \+(G)

  This code is based on the famous Prolog benchmark flatten.pl.
  This code uses dynamic predicates disj_counter/1 and filename/1.
*/
eliminate_disj(Cl, NewCl, DummyCls) :-
	extract_disj(Cl, NewCl, Disjs, []),
	treat_disj(Disjs, DummyCls, []).

extract_disj(Cl, Cl) --> {var(Cl)}, !.
extract_disj((H :- B), (H :- NewB)) --> !,
	disj_extraction(B, NewB, (H :- B)).
extract_disj(Cl, Cl) --> !.

disj_extraction(G, G, _) --> {var(G)}, !.
disj_extraction((G1 & G2), (GG1 & GG2), Cl) --> !,
	disj_extraction(G1, GG1, Cl),
	disj_extraction(G2, GG2, Cl).
disj_extraction((G1, G2), (GG1, GG2), Cl) --> !,
	disj_extraction(G1, GG1, Cl),
	disj_extraction(G2, GG2, Cl).
disj_extraction((R -<> G), (RR -<> GG), Cl) --> !,
	disj_extraction_R(R, RR),
	disj_extraction(G, GG, Cl).
disj_extraction((R => G), (RR => GG), Cl) --> !,
	disj_extraction_R(R, RR),
	disj_extraction(G, GG, Cl).
disj_extraction( !G, !GG, Cl) --> !,
	disj_extraction(G, GG, Cl).
disj_extraction(G, GG, Cl) --> {is_disj(G, G1)}, !,
	{inc_disj_counter(ID)},
	[disj(G1, ID, GG, Cl)].
disj_extraction(G, G, _) --> !.

disj_extraction_R(R, R) --> {var(R)}, !.
disj_extraction_R((R1, R2), (RR1, RR2)) --> !,
	disj_extraction_R(R1, RR1),
	disj_extraction_R(R2, RR2).
disj_extraction_R(R, RR) -->
	disj_extraction_S(R, RR).

disj_extraction_S(R, R) --> {var(R)}, !.
disj_extraction_S((R1 & R2), (RR1 & RR2)) --> !,
	disj_extraction_S(R1, RR1),
	disj_extraction_S(R2, RR2).
disj_extraction_S((forall Xs\ G -<> A), (forall Xs\ G1 -<> A1)) --> !,
	extract_disj((A :- G), Cl),
	{Cl = (A1 :- G1)}.
disj_extraction_S((G -<> A), (G1 -<> A1)) --> !,
	extract_disj((A :- G), Cl),
	{Cl = (A1 :- G1)}.
disj_extraction_S(R, R) --> !.

is_disj((C1 -> C2), ((C1, !, C2) ; fail)) :- !.
is_disj(((C1 -> C2) ; C3), ((C1, !, C2) ; C3)) :- !.
is_disj((C1 ; C2), (C1 ; C2)) :- !.
is_disj(not(C),((C, !, fail) ; true)) :- !.
is_disj(\+(C),((C, !, fail) ; true)).

inc_disj_counter(N) :- 
	retract(disj_counter(N)),
	!,
	N1 is N+1,
	assert(disj_counter(N1)).
inc_disj_counter(0) :- 
	llpc_error([dynamic,predicate,disj_counter/1,is,not,asserted,yet]),
	assert(disj_counter(0)).

treat_disj([]) --> !.
treat_disj([disj((A ; B), ID, NewG, Cl)|Ds]) -->
	{variables((A ; B), V1)},
	{variables(Cl, V2)},
	{intersect_vars(V1, V2, Args)},
	{dummy_name(ID, Name)},
	{NewG =.. [Name|Args]},
	dummy_clauses((A ; B), NewG),
	treat_disj(Ds).

intersect_vars(V1,V2,Out) :-
	sort(V1,Sorted1),
	sort(V2,Sorted2),
	intersect_sorted_vars(Sorted1,Sorted2,Out).

intersect_sorted_vars([],_,[]) :- !.
intersect_sorted_vars(_,[],[]).
intersect_sorted_vars([X|Xs],[Y|Ys],[X|Rs]) :- X == Y, !,
	intersect_sorted_vars(Xs,Ys,Rs).
intersect_sorted_vars([X|Xs],[Y|Ys],Rs) :- X @< Y, !,
	intersect_sorted_vars(Xs,[Y|Ys],Rs).
intersect_sorted_vars([X|Xs],[Y|Ys],Rs) :- X @> Y, !,
	intersect_sorted_vars([X|Xs],Ys,Rs).

dummy_name(ID, Name) :- filename(File), !,
	list_to_string(['$dummy_', File, '_', ID], Name).
dummy_name(ID, Name) :-
	llpc_error([dynamic,predicate,filename/1,is,not,asserted,yet]),
	list_to_string(['$dummy_unknown', '_', ID], Name),
	assert(filename(unknown)).

dummy_clauses((A;B), X) --> !,
	dummy_clauses(A,X),
	dummy_clauses(B,X).
dummy_clauses(A, X) -->
	{copy_term((X :- A), DummyCl)},
	[DummyCl].


