/****************************************************************
	LLPAM Emulator
	(debugging)
	by Naoyuki Tamura (tamura@kobe-u.ac.jp)
	   Eiji Sugiyama
	   Yuichi Ikeda
	   Dept. CS, Kobe University
	and
	Mutsunori Banbara (banbara@cs32.scitec.kobe-u.ac.jp)
	Last modified on 30 Julu 2000 by M.Banbara
****************************************************************/
#include <stdio.h>

#define EXTERN extern
#include "llpam.h"
#include "debug.h"
#define DECLARE_OP_TABLE
#include "instr.h"

void am_print_functor(CELL x)
{
    if (TAG(x) != TagCON) {
	printf("Internal error in am_print_functor\n");
	am_abort(201);
    }
    printf("%s", SYMBOL(CON2SYM(x)).print_name);
}


void am_print_functor_arity(CELL x)
{
    am_print_functor(x);
    printf("/%d", SYMBOL(CON2SYM(x)).arity);
}


void am_print_term(CELL x)
{
    STOREPTR p;
    int n, i, t;

    if (x == UNDEF){
	printf("UNDEF");
	return;
    }
    AM_DEREF(x, t, x);
    switch (t) {
      case TagREF:
	printf("_%ld", PTR2INX(REF2PTR(x)) - PTR2INX(heap_top));
	break;
      case TagINT:
	printf("%ld", INT2LONG(x));
	break;
      case TagCON:
	am_print_functor(x);
	break;
      case TagLIS:
	printf("[");
	for (;;) {
	    p = LIS2PTR(x);
	    am_print_term(STORE(p));
	    AM_DEREF(x, t, STORE(p+1));
	    if (t == TagLIS) {
		printf(",");
		continue;
	    }
	    if (x == SYM_NIL) {
		break;
	    }
	    printf("|");
	    am_print_term(x);
	    break;
	}
	printf("]");
	break;
      case TagSTR:
	p = STR2PTR(x);
	am_print_functor(STORE(p));
	n = SYMBOL(CON2SYM(STORE(p))).arity;
	if (n > 0) {
	    printf("(");
	    for (i = 1; i <= n; i++) {
		am_print_term(STORE(p+i));
		if (i < n)
		  printf(",");
	    }
	    printf(")");
	}
	break;
      case TagCLO:
	p = CLO2PTR(x);
	printf("<CLO, %ld> [", PTR2INX(p));
	n = INT2LONG(STORE(p+1));
	if (n > 0){
	    for (i = 1; i <= n; i++) {
		am_print_term(STORE(p+1+i));
		if (i < n)
		    printf(",");
	    }
	}
	printf("]");
	break;
      case TagRES:
	  printf("%ld", RES2RESINX(x));
	  break;
      case TagTOP:
	  printf("%ld", TOP2BOOL(x));
	  break;
    }
}


void am_dump_cell(CELL x)
{
    if (x == UNDEF){
	printf("UNDEF");
	return;
    }
    switch (TAG(x)) {
      case TagREF:
	printf("<REF, %ld>", PTR2INX(REF2PTR(x)));
	if (x != STORE(REF2PTR(x))) {
	    printf(" -> ");
	    am_dump_cell(STORE(REF2PTR(x)));
	}
	break;
      case TagINT:
	printf("<INT, %ld>", INT2LONG(x));
	break;
      case TagCON:
	printf("<CON, %ld>", CON2SYM(x));
	printf(" = ");
	am_print_term(x);
	break;
      case TagLIS:
	printf("<LIS, %ld>", PTR2INX(LIS2PTR(x)));
	printf(" = ");
	am_print_term(x);
	break;
      case TagSTR:
	printf("<STR, %ld>", PTR2INX(STR2PTR(x)));
	printf(" = ");
	am_print_term(x);
	break;
      case TagCLO:
	printf("<CLO, %ld>", PTR2INX(CLO2PTR(x)));
	break;
      case TagRES:
	printf("<RES, %ld>", RES2RESINX(x));
	break;
      case TagTOP:
	  printf("%ld", TOP2BOOL(x));
	  break;
    }
}


void am_dump_ptr(CELL x)
{
    CELL y;
    STOREPTR p;
    int n, i, t;

    AM_DEREF(y, t, x);
    switch (t) {
      case TagLIS:
	p = LIS2PTR(y);
	printf("<LIS, %ld>\n", PTR2INX(p));
	printf("\t");
	am_dump_cell(STORE(p));
	printf("\n");
	printf("\t");
	am_dump_cell(STORE(p+1));
	printf("\n");
	am_dump_ptr(STORE(p));
	am_dump_ptr(STORE(p+1));
	break;
      case TagSTR:
	p = STR2PTR(y);
	printf("<STR, %ld>\n", PTR2INX(p));
	printf("\t");
	am_print_functor_arity(STORE(p));
	printf("\n");
	n = SYMBOL(CON2SYM(STORE(p))).arity;
	for (i = 1; i <= n; i++) {
	    printf("\t");
	    am_dump_cell(STORE(p+i));
	    printf("\n");
	}
	for (i = 1; i <= n; i++)
	  am_dump_ptr(STORE(p+i));
	break;
      case TagCLO:
	p = CLO2PTR(x);
	printf("<CLO, %ld\n>", PTR2INX(p));
	n = INT2LONG(STORE(p+1));
	printf("\t");
	printf("num_of_freevars = %d\n", n);
	printf("\t[");
	if (n > 0){
	    for (i = 1; i <= n; i++) {
		am_print_term(STORE(p+1+i));
		if (i < n)
		    printf(",");
	    }
	}
	printf("]");
	break;
      default:
	break;
    }
}


void am_dump_term(CELL x)
{
    am_dump_cell(x);
    printf("\n");
    am_dump_ptr(x);
    printf("\n");
}


int am_list_instr(STOREPTR p)
{
    STOREPTR oldP;
    int n;

    oldP = P;
    P = p;
    if (OP >= LAST_OP) {
	printf("%6ld:\tIllegal instruction\n", PTR2INX(P));
	P = oldP;
	return(-1);
    }
    printf ("%6ld:\t%-18s", PTR2INX(P), op_table[OP].op_name);
    switch (op_table[OP].op_type) {
      case  0: /* OP */
	n = 1;
	break;
      case  1: /* OP+ARG1 */
	n = 1;
	printf("%d", CARG1);
	break;
      case  2: /* OP+ARG1+ARG2 */
	n = 1;
	printf("%d, %d", CARG1, CARG2);
	break;
      case  3: /* OP+ARG1, f/n */
      case  4: /* OP+ARG1, c */
	n = 2;
	am_print_functor_arity(CARGW(1));
	printf(", %d", CARG1);
	break;
      case  5: /* OP+ARG1, i */
	n = 2;
	printf("%ld, %d", INT2LONG(CARGW(1)), CARG1);
	break;
      case  6: /* OP, f/n */
      case  7: /* OP, c */
	n = 2;
	am_print_functor_arity(CARGW(1));
	break;
      case  8: /* OP, i */
	n = 2;
	printf("%ld", INT2LONG(CARGW(1)));
	break;
      case  9: /* OP, L */
	n = 2;
	printf("%ld", CARGW(1));
	break;
      case 10: /* OP, L1,L2,L3,L4,L5 */
	n = 6;
	printf("%ld, ", CARGW(1));
	printf("%ld, ", CARGW(2));
	printf("%ld, ", CARGW(3));
	printf("%ld, ", CARGW(4));
	printf("%ld"  , CARGW(5));
	break;
      case 11: /* OP+ARG1+ARG2, ARG3 */
	n = 2;
	printf("%d, %d, ", CARG1, CARG2);
	printf("%d", CARG3);
	break;
      case 12: /* OP+ARG1+ARG2, ARG3+ARG4 */
	n = 2;
	printf("%d, %d, ", CARG1, CARG2);
	printf("%d, %d", CARG3, CARG4);
	break;
      case 13: /* OP+ARG1+ARG2, ARG3+ARG4+ARG5 */
	n = 2;
	printf("%d, %d, ", CARG1, CARG2);
	printf("%d, %d, %d", CARG3, CARG4, CARG5);
	break;
      case 14: /* OP+ARG1, Term */
	n = 2;
	printf("0x%08lx, %d \% ", CARGW(1), CARG1);
	am_print_term(CARGW(1));
	break;
      case 15: /* OP, Term */
	n = 2;
	printf("0x%08lx \% ", CARGW(1));
	am_print_term(CARGW(1));
	break;
      case 16: /* OP+ARG1, L, i */
	n = 3;
	printf("%ld, ", CARGW(1));
	printf("%ld  ", INT2LONG(CARGW(2)));
	printf("%d"   , CARG1);
	break;
      case 17: /* OP+ARG1, f/n, L */
	n = 3;
	am_print_functor_arity(CARGW(1));
	printf(", %d, ", CARG1);
	printf("%ld"   , CARGW(2));
	break;
      case 18: /* OP+ARG1+ARG2, L */
	n = 2;
	printf("%d, %d, ", CARG1, CARG2);
	printf("%ld"   , CARGW(1));
	break;
    }
    printf("\n");
    P = oldP;
    return(n);
}


void am_trace_goal(CELL goal)
{
    SYMBOLINX pred;
    int t, notrace;

    AM_DEREF(goal, t, goal);
    notrace = NO;
    if (t == TagCON) {
	pred = CON2SYM(goal);
	if (SYMBOL(pred).plevel >= 1 ||
	    *(SYMBOL(pred).print_name) == '$')
	  notrace = YES;
    } else if (t == TagSTR) {
	pred = CON2SYM(STORE(STR2PTR(goal)));
	if (SYMBOL(pred).plevel >= 1 ||
	    *(SYMBOL(pred).print_name) == '$')
	  notrace = YES;
    }
    if (notrace) {
	am_exec_indirect(goal);
	return;
    }
    predicate = SYM_TRACE;
    num_of_args = 1;
    A(1) = goal;
    P = SYMBOL(CON2SYM(predicate)).codeaddr;
    if (P == PRED_UNDEF) {
	if (prolog_flag.unknown == 0) {
	    printf("%s/%d is undefined\n",
		   SYMBOL(CON2SYM(predicate)).print_name, num_of_args);
	    am_abort(11);
	}
	AM_FAIL;
	return;
    }	
    B0 = B;
}


/*
 Local Variables:
 mode: c
 c-indent-level: 4
 End:
*/
