/**
 * @(#)Peg.java
 * @author Naoyuki Tamura (tamura@kobe-u.ac.jp)
 * Dept. CS, Kobe University
 *
 * Peg applet
 */
import java.awt.*;
import java.io.*;
import java.util.*;

public class Peg extends PrologFrame {
    private PegPanel pegPanel;
    private TextField N_Field;
    private int N = 5;
    private int I0 = 3;
    private int J0 = 2;

    public Peg() {
	setSize(500, 400);
	setLayout(new BorderLayout());
	add("North", new Label("Peg", Label.CENTER));
	pegPanel = new PegPanel();
	add("Center", pegPanel);
	Panel p = new Panel();
	add("South", p);
	p.add(new Label("N="));
	N_Field = new TextField(Integer.toString(N), 2);
	p.add(N_Field);
	control = new PrologControlPanel("llp -ejava peg");
	p.add(control);
	validate();
    }

    void sendArgument() {
	String s;
	pegPanel.stop();	// stop animation
	try {
	    N = Integer.parseInt(N_Field.getText());
	} catch (NumberFormatException e) {
	    N = 1;
	    N_Field.setText(Integer.toString(N));
	}
	s = Integer.toString(N)  + "," + 
	    Integer.toString(I0) + "," +
	    Integer.toString(J0) + ".";
	control.sendArgument(s);
    }

    void receiveResult() {
	pegPanel.stop();	// stop animation
	String result = control.receiveResult();
	if (result == null || result.equals("DONE")) {
	    return;
	}
	StringTokenizer st = new StringTokenizer(result, ",");
	try {
	    int len = Integer.parseInt(st.nextToken());
	    int moves[][] = new int[len][4];
	    for (int k = 0; k < len; ++k) {
		moves[k][0] = Integer.parseInt(st.nextToken());
		moves[k][1] = Integer.parseInt(st.nextToken());
		moves[k][2] = Integer.parseInt(st.nextToken());
		moves[k][3] = Integer.parseInt(st.nextToken());
	    }
	    pegPanel.setSol(N, I0, J0, moves);
	} catch (NumberFormatException e) {}
	pegPanel.start();	// start animation
    }

    void aborted() {
	pegPanel.stop();	// stop animation
    }

    public static void main(String args[]) {
	new Peg();
    }
}


class PegPanel extends Panel implements Runnable {
    private Thread pegAnimator = null;
    private int N;
    private int I0;
    private int J0;
    private Color peg[][] = null;
    private int moves[][] = null;
    private int cellSize;
    private int x0;
    private int y0;
    private Graphics g;
    private int wait = 300;

    public synchronized void start() {
	stop();
	pegAnimator = new Thread(this);
	pegAnimator.start();
    }

    public synchronized void stop() {
	if (pegAnimator != null)
	    pegAnimator.stop();
	pegAnimator = null;
    }

    public synchronized void setSol(int n, int i0, int j0, int m[][]) {
	N = n;
	I0 = i0;
	J0 = j0;
	moves = m;
	peg = null;
    }

    private void drawPeg(Color c, int i, int j) {
	g.setColor(c);
	--i;
	--j;
	int x = x0 + (N-1-i)*cellSize/2 + j*cellSize;
	int y = y0 + i*cellSize;
	g.fillOval(x+cellSize/4, y+cellSize/4, cellSize/2, cellSize/2);
    }

    public void paint(Graphics g) {
	if (peg == null)
	    return;
	this.g = g;
	int w = getSize().width;
	int h = getSize().height;
	cellSize = Math.min(w, h) / N;
	x0 = (w - N * cellSize) / 2;
	y0 = (h - N * cellSize) / 2;
	for (int i = 1; i <= N; ++i)
	    for (int j = 1; j <= i; ++j)
		drawPeg(peg[i][j], i, j);
    }

    private void sleep(int t) {
	try {
	    Thread.sleep(t);
	} catch (InterruptedException e) {}
    }

    public void run() {
	if (moves == null)
	    return;
	peg = new Color[N+1][N+1];
	for (int i = 1; i <= N; ++i)
	    for (int j = 1; j <= i; ++j)
		peg[i][j] = Color.blue;
	peg[I0][J0] = Color.white;
	repaint();
	for (int k = 0; k < moves.length; ++k) {
	    sleep(wait);
	    int i0 = moves[k][0];
	    int j0 = moves[k][1];
	    int di = moves[k][2];
	    int dj = moves[k][3];
	    peg[i0][j0] = Color.red;
	    repaint();
	    sleep(wait);
	    peg[i0][j0] = Color.white;
	    peg[i0+di][j0+dj] = Color.white;
	    peg[i0+2*di][j0+2*dj] = Color.blue;
	    repaint();
	    sleep(wait);
	}
    }

}


