:- op(1060, xfy, (&)).
:- op( 950, xfy, (-<>)).
:- op( 950, xfy, (=>)).
:- op( 900,  fy, (!)).

ex1 :-
	prove((p=>((q&r)-<>(p,q)))).

ex2 :-
	prove(((p&q)-<>(p&q))).

ex3 :-
	prove((p-<>q-<>r-<>(p,erase,r))).

ex4 :-
	prove((p-<>q-<>r-<>(p,q,erase))).

ex5 :-
	prove((p-<>q-<>r-<>(p,q))).

prove(G) :-
	I = [], O = [], L = 1, U = -1,
	prove(G, I, O, L, U).

prove(fail, _, _, _, _) :- !, fail.
prove(true, I, I, L, U) :- !.
prove(erase, I, O, L, U) :- !,
	subcontext(O, I, U, L).
prove((G1,G2), I, O, L, U) :- !,
	prove(G1, I, M, L, U),
	prove(G2, M, O, L, U).
prove((G1&G2), I, O, L, U) :- !,
	L1 is L+1, U1 is U-1,
	prove(G1, I, M, L, U1),
	change(M, N, U1, L1),
	prove(G2, N, O, L1, U),
	thinable(O, L1).
prove((G1;G2), I, O, L, U) :- !,
	(prove(G1, I, O, L, U) ;
	 prove(G2, I, O, L, U)).
prove(!G, I, O, L, U) :- !,
	L1 is L+1, U1 is U-1,
	prove(G, I, O, L1, U1).
prove((R -<> G), I, O, L, U) :- !,
	prove(G, [(R,L)|I], [(R,U)|O], L, U).
prove((R => G), I, O, L, U) :- !,
	prove(G, [(R,0)|I], [(R,0)|O], L, U).
prove(A, I, O, L, U) :- % A is an atomic formula
	pick(I, O, A, L, U).
prove(A, I, O, L, U) :- % A is an atomic formula
	pick(I, M, (G -<> A), L, U),
	prove(G, M, O, L, U).

subcontext([], [], L, U).
subcontext([(R,L)|I], [(R,U)|O], L, U) :-
	subcontext(I, O, L, U).
subcontext([X|I], [X|O], L, U) :-
	subcontext(I, O, L, U).

change([], [], L1, L2) :- !.
change([(R,L1)|I], [(R,L2)|O], L1, L2) :- !,
	change(I, O, L1, L2).
change([(R,L)|I], [(R,L)|O], L1, L2) :- L=\=L1, !,
	change(I, O, L1, L2).

thinable([], L) :- !.
thinable([(R,L1)|I], L) :- L1=\=L,
	thinable(I, L).

%%%
pick(I, I, S, L, U) :-
	builtin(S),
	!,
	call(S).
%%%
pick(I, I, S, L, U) :-
	rule(S).
pick([(R,0)|I], [(R,0)|I], S, L, U) :-
	pick_S(R, S).
pick([(R,L)|I], [(R,U)|I], S, L, U) :-
	pick_S(R, S).
pick([X|I], [X|O], S, L, U) :-
	pick(I, O, S, L, U).

pick_S((R1&R2), S) :- !,
	(pick_S(R1, S); pick_S(R2, S)).
pick_S(S, S).

builtin(write(_)) :- !.
builtin(writeq(_)) :- !.
builtin(read(_)) :- !.
builtin(nl) :- !.
builtin(tab(_)) :- !.
builtin(var(_)) :- !.
builtin(nonvar(_)) :- !.
builtin(atomic(_)) :- !.
builtin(integer(_)) :- !.
builtin(atom(_)) :- !.
builtin((_ = _)) :- !.
builtin((_ == _)) :- !.
builtin((_ =:= _)) :- !.
builtin((_ =\= _)) :- !.
builtin((_ > _)) :- !.
builtin((_ =< _)) :- !.
builtin((_ < _)) :- !.
builtin((_ >= _)) :- !.
builtin((_ is _)) :- !.
builtin(compare(_,_,_)) :- !.
builtin(functor(_,_,_)) :- !.
builtin(arg(_,_,_)) :- !.
builtin((_ =.. _)) :- !.

%rule( (append(X, Y, [1,2,3,4,5]), write([X,Y]),nl,fail) -<> main).
%rule(  main ).
%rule( append([], Zs, Zs) ).
%rule(( append(Xs, Ys, Zs) -<> append([X|Xs], Ys, [X|Zs]) )).

%rule((queen(N, Q), write(Q), nl, fail) -<> queen_solve(N)).
%rule(queen_solve(_)).
rule((n(N) -<> result(Q) -<> place(N)) -<> queen(N, Q)).
rule((n(N), (c(1) -<>u(2) -<> d(0) -<> solve(N, []))) -<> place(1)).
rule((I > 1, I1 is I-1, U1 is 2*I, U2 is 2*I-1, D1 is I-1, D2 is 1-I,(c(I)-<>u(U1)-<>u(U2)-<>d(D1)-<>d(D2) -<> place(I1))) -<> place(I)).
rule((result(Q), erase) -<> solve(0, Q)).
rule((I > 0, c(J), U is I+J, u(U), D is I-J, d(D), I1 is I-1, solve(I1, [J|Q])) -<> solve(I, Q)).


%rule( (test(N), fail) -<> main ).
%rule( main ).
%rule(((p(1) -<> p(2) -<> (p(X), p(Y))), write([X,Y]), nl) -<> test(1)).
%rule(((p(1) -<> p(2) -<> ((p(X) & p(Y)), p(Z))), write([X,Y,Z]), nl) -<> test(2)).
%rule((((p(1) & p(2)) -<> p(X)), write([X]), nl) -<> test(3)).
%rule(((p(1) => (p(X), p(Y))), write([X,Y]), nl) -<> test(4)).
%rule(((p(1) => p(2) => (p(X), p(Y))), write([X,Y]), nl) -<> test(5)).
%rule(((p(1) => p(2) -<> ((p(X) & p(Y)), p(Z))), write([X,Y,Z]), nl) -<> test(6)).

