:- op(1060, xfy, (&)).
:- op( 950, xfy, [-<>, =>]).
:- op( 900,  fy, [!, @, #]).


prove(G) :-
	I = [], O = [], L = 1, U = -1, T = 1,
	prove(G, I, O, L, U, T).

prove(fail, _, _, _, _, _) :- !, fail.
prove(true, I, I, _L, _U, _T) :- !.
prove(erase, I, O, L, U, T) :- !,
	subcontext(O, I, T, U, L).
prove((G1,G2), I, O, L, U, T) :- !,
	prove(G1, I, M, L, U, T),
	prove(G2, M, O, L, U, T).
prove((G1&G2), I, O, L, U, T) :- !,
	L1 is L+1, U1 is U-1,
	prove(G1, I, M, L, U1, T),
	change(M, N, U1, L1),
	prove(G2, N, O, L1, U, T),
	thinable(O, L1).
prove((G1;G2), I, O, L, U, T) :- !,
	(prove(G1, I, O, L, U, T) ;
	 prove(G2, I, O, L, U, T)).
prove(!G, I, O, L, U, T) :- !,
	L1 is L+1, U1 is U-1,
	prove(G, I, O, L1, U1, T).
prove(@G, I, O, L, U, T) :- !,
	T1 is T+1,
	prove(G, I, O, L, U, T1).
prove((R -<> G), I, O, L, U, T) :- !,
        count_next(R, N, R1),
        T1 is T + N,
	prove(G, [(R1,L,T1)|I], [(R1,U,T1)|O], L, U, T).
prove((S => G), I, O, L, U, T) :- !,
	prove(G, [(S,0,0)|I], [(S,0,0)|O], L, U, T).
prove(A, I, O, L, U, T) :- % A is an atomic formula
	pick(I, O, A, L, U, T).
prove(A, I, O, L, U, T) :- % A is an atomic formula
	pick(I, M, (G -<> A), L, U, T),
	prove(G, M, O, L, U, T).

count_next(@R, N, R1) :- !,
        count_next(R, N1, R1),
        N is N1 + 1.
count_next(R, 0, R).

subcontext([], [], _T, _U, _L).
subcontext([(S,0,0)|O], [(S,0,0)|I], T, U, L) :-
	subcontext(O, I, T, U, L).
subcontext([(#R,U,T0)|O], [(#R,L,T0)|I], T, U, L) :-
	subcontext(O, I, T, U, L).
subcontext([(R,U,T0)|O], [(R,L,T0)|I], T, U, L) :-
	\+(R = (#_)),
        T0 >= T,
	subcontext(O, I, T, U, L).
subcontext([X|O], [X|I], T, U, L) :-
	subcontext(O, I, T, U, L).

change([], [], _L1, _L2) :- !.
change([(R,L1,T)|I], [(R,L2,T)|O], L1, L2) :- !,
	change(I, O, L1, L2).
change([(R,L,T)|I], [(R,L,T)|O], L1, L2) :- L=\=L1, !,
	change(I, O, L1, L2).

thinable([], _L) :- !.
thinable([(_R,L1,_T)|I], L) :- L1=\=L,
	thinable(I, L).

%%%
pick(I, O, S, _L, _U, _T) :-
	pick1(I, O, S, _L, _U, _T).
pick(I, I, S, _L, _U, _T) :-
	builtin(S),
	!,
	call(S).
pick(I, I, S, _L, _U, _T) :-
	rule(S).
pick(I, I, (G -<> A), _L, _U, _T) :-
        rule((A :- G)).

pick1([(S,0,0)|I], [(S,0,0)|I], S, _L, _U, _T).
pick1([(#R,L,T0)|I], [(#R,U,T0)|I], S, L, U, T) :-
	T >= T0,
	pick_S(R, S).
pick1([(R,L,T)|I], [(R,U,T)|I], S, L, U, T) :-
        \+(R = (#_)),
	pick_S(R, S).
pick1([X|I], [X|O], S, L, U, T) :-
	pick1(I, O, S, L, U, T).

pick_S((R1&R2), S) :- !,
	(pick_S(R1, S); pick_S(R2, S)).
pick_S(S, S). 

builtin(write(_)) :- !.
builtin(writeq(_)) :- !.
builtin(read(_)) :- !.
builtin(nl) :- !.
builtin(tab(_)) :- !.
builtin(var(_)) :- !.
builtin(nonvar(_)) :- !.
builtin(atomic(_)) :- !.
builtin(integer(_)) :- !.
builtin(atom(_)) :- !.
builtin((_ = _)) :- !.
builtin((_ == _)) :- !.
builtin((_ =:= _)) :- !.
builtin((_ =\= _)) :- !.
builtin((_ > _)) :- !.
builtin((_ =< _)) :- !.
builtin((_ < _)) :- !.
builtin((_ >= _)) :- !.
builtin((_ is _)) :- !.
builtin(compare(_,_,_)) :- !.
builtin(functor(_,_,_)) :- !.
builtin(arg(_,_,_)) :- !.
builtin((_ =.. _)) :- !.

rule(( p(V,V,[V]) :- v(V) )).
rule(( p(U,V,[U|P]) :-
         v(U), e(U,W), @p(W,V,P) )).
rule(( e(_U,_V) )).
rule(( goal(P) :- #v(a) -<> @ @v(b) -<>
         @ #v(c) -<> #v(d) -<> p(a,d,P) )).

rule( append([], Zs, Zs) ).
rule(( append([X|Xs], Ys, [X|Zs]) :- append(Xs, Ys, Zs) )).