/*
  TLLP --> LLP translator
  by Mutsunori Banbara (banbara@cs32.scitec.kobe-u.ac.jp)
     Nara National College of Technology
*/

:- dynamic internal_clause/2, internal_predicate/2.

:- op(1200, fx,  (forall)).
:- op(1190, xfx, (\)).
:- op(1170, xfx, (<=)).
:- op(1170, xfx, (:-)).
:- op(1170, xfx, (-->)).
:- op(1170,  fx, (:-)).
:- op(1170,  fx, (?-)).
:- op(1150,  fx, (resource)).
:- op(1060, xfy, (&)).
:- op( 950, xfy, (-<>)).
:- op( 950, xfy, (=>)).
:- op( 900,  fy, (!)).
:- op( 900,  fy, (@)).
:- op( 900,  fy, (#)).

main :-
	read(X),
	tllpc(X).

tllpc([TLLP, LLP]) :-
	tllpc(TLLP, LLP).
tllpc(_).

tllpc(TLLP, LLP) :-
	read_program(TLLP),
	open(LLP, write, Out),
	translate_program(Out),
	close(Out).

/**************************************************************
	Read progam
**************************************************************/
read_program(TLLP) :-
	retractall(internal_clause(_,_)),
	retractall(internal_predicate(_,_)),
	open(TLLP, read, In),
	repeat,
	  read_stream(In, X),
	  assert_clause(X),
	X == end_of_file,
	close(In),
	!.

assert_clause(end_of_file) :- !.
assert_clause(Clause) :-
	assert_cls(Clause).

assert_cls((Head :- Body)) :- !,
	assert_predicate(Head),
	assert(internal_clause(Head, Body)).
assert_cls(Head) :- !,
	assert_predicate(Head),
	assert(internal_clause(Head, true)).

assert_predicate(Head) :-
	functor(Head, Functor, Arity),
	clause(internal_predicate(Functor, Arity), _),
	!.
assert_predicate(Head) :-
	functor(Head, Functor, Arity),
	assert(internal_predicate(Functor, Arity)).

read_stream(In, X) :-
	current_input(Stream),
	set_input(In),
	read(X),
	set_input(Stream).

/**************************************************************
	Translate Program
**************************************************************/
translate_program(Out) :-
	clause(internal_predicate(Functor, Arity), _),
	functor(Head, Functor, Arity),
	findall((Head :- Body), internal_clause(Head, Body), Clauses0),
	translate_predicate(Clauses0, Clauses),
	write_clauses(Out, Clauses),
	nl(Out),
	fail.
translate_program(_).

write_clauses(Out, Clauses) :-
	current_output(X),
	set_output(Out),
	write_clauses(Clauses),
	set_output(X).

write_clauses([]) :- !.
write_clauses([C|Cs]) :- 
	write_cls(C),
	write_clauses(Cs).

write_cls(C) :-
    numbervars(C, 0, _), 
    write_cls0(C).

write_cls0((H :- B)) :- B == true, !,
    writeq(H),write('.'),nl.
write_cls0((H :- B)) :-
    writeq(H), write(' :- '), nl,
    write_body(B).

write_body(((B1,B2),B3)) :- !,
	write_body((B1,(B2,B3))).
write_body((B1,B2)) :- !,
	tab(8), writeq(B1), write(','), nl,
	write_body(B2).
write_body((B)) :-
	tab(8), writeq(B), write('.'), nl.

/**************************************************************
	Translate Clause
**************************************************************/
translate_predicate([], []) :- !.
translate_predicate([C|Cls], [C1|Cls1]) :- 
	trans_clause(C, C1),
	translate_predicate(Cls, Cls1).

trans_clause(Clause, NewClause) :- 
	trans_clause0(Clause, Clause1),
	NewClause = Clause1.

trans_clause0(Cl, Cl) :- var(Cl), !.
trans_clause0(Cl, Cl1) :-  
	trans_cls(Cl, Cl1), !.
trans_clause0(Cl, Cl).


trans_cls((H :- B), (H1 :- B1)) :- !,
	trans_head(H, T, H1),
	trans_body(B, T, B1).
trans_cls(H, H1) :-
	trans_head(H, _T, H1).

trans_head(H, T, H1) :-
	add_time(H, T, H1).

trans_body(G, _, _) :- var(G), !,
	tllpc_message([goal,'(',G,')',must,be,instantiated]),
	fail.
trans_body(true, _T, true) :- !.
trans_body(false, _T, false) :- !.
trans_body(fail, _T, fail) :- !.
trans_body(top, _T, top) :- !.
trans_body(erase, _T, erase) :- !.
trans_body((G1, G2), T, (GG1, GG2)) :- !,
	trans_body(G1, T, GG1),
	trans_body(G2, T, GG2).
trans_body((G1 & G2), T, (GG1 & GG2)) :- !,
	trans_body(G1, T, GG1),
	trans_body(G2, T, GG2).
trans_body((IF0 -> THEN0), T, (IF -> THEN)) :- !,
	trans_body(IF0, T, IF),
	trans_body(THEN0, T, THEN).
trans_body((IF0 -> THEN0 ; ELSE0), T, (IF -> THEN ; ELSE)) :- !,
	trans_body(IF0, T, IF),
	trans_body(THEN0, T, THEN),
	trans_body(ELSE0, T, ELSE).
trans_body((G1 ; G2), T, (GG1 ; GG2)) :- !,
	trans_body(G1, T, GG1),
	trans_body(G2, T, GG2).
trans_body((R -<> G), T, GG) :- !,
	trans_res_R(R, T, R1, Extra),
	trans_body(G, T, G1),
	(var(Extra) ->
	     GG = (R1 -<> G1)
	  ;
	     GG = (Extra, (R1 -<> G1))
	).
trans_body((S => G), T, (SS => GG)) :- !,
	trans_res_S3(S, T, SS),
	trans_body(G, T, GG).
trans_body((!G), T, (!GG)) :- !,
	trans_body(G, T, GG).
trans_body((@G), T, GG) :- !,
	count_next((@G), N, G1),
	trans_body(G1, T1, G2),
	GG = (T1 is T+N, G2).
trans_body(G, _, _) :- G = (forall _\ _), !,
	tllpc_message([goal,'(',G,')',is,invalid]),
	fail.
trans_body(G, _, _) :- G = (#_), !,
	tllpc_message([goal,'(',G,')',is,invalid]),
	fail.
trans_body(G, T, (\+ GG)) :- G = (\+ G1), !,
	trans_body(G1, T, GG).
trans_body(!, _, !) :- !.
trans_body(G, _, G) :-
	isbuiltin(G), !.
trans_body(G, T, G1) :-
	add_time(G, T, G1).

isbuiltin(write(_)).
isbuiltin(writeq(_)).
isbuiltin(read(_)).
isbuiltin(nl).
isbuiltin(tab(_)).
isbuiltin(var(_)).
isbuiltin(nonvar(_)).
isbuiltin(atomic(_)).
isbuiltin(integer(_)).
isbuiltin(atom(_)).
isbuiltin((_ = _)).
isbuiltin((_ == _)).
isbuiltin((_ =:= _)).
isbuiltin((_ =\= _)).
isbuiltin((_ > _)).
isbuiltin((_ =< _)).
isbuiltin((_ < _)).
isbuiltin((_ >= _)).
isbuiltin((_ is _)).
isbuiltin(compare(_,_,_)).
isbuiltin(functor(_,_,_)).
isbuiltin(arg(_,_,_)).
isbuiltin((_ =.. _)).
isbuiltin(!).

trans_res_R(R, _, _,_) :- var(R), !,
	tllpc_message([resource,'(',R,')',must,be,instantiated]),
	fail.	
trans_res_R(R, T, RR, Extra) :-
	count_next(R, N, R1),
	trans_res_R0(R1, T+N, RR, Extra).

trans_res_R0(R, _, _, _) :- var(R), !,
	tllpc_message([resource,'(',R,')',must,be,instantiated]),
	fail.	
trans_res_R0((#R), T, RR, _) :- !,
	trans_res_S2(R, T, RR).
trans_res_R0(R, T, RR, Extra) :- 
	trans_res_S1(R, T, RR, Extra).

trans_res_S1(R, T+N, R1, _) :- N =:= 0, !,
	trans_res_S1a(R, T, R1).
trans_res_S1(R, T, R1, (T1 is T)) :- 
	trans_res_S1a(R, T1, R1).

trans_res_S1a(R, _, _) :- var(R), !,
	tllpc_message([resource,'(',R,')',must,be,instantiated]),
	fail.	
trans_res_S1a((R1&R2), T, (RR1&RR2)) :- !,
	trans_res_S1a(R1, T, RR1),
	trans_res_S1a(R2, T, RR2).
trans_res_S1a((forall Xs\ S), T, (forall Xs\ SS)) :- !,
	trans_res_S1a(S, T, SS).
trans_res_S1a((G -<> H), T, (G1 -<> H1)) :- resource_head(H), !,
	add_time(H, T, H1),
	trans_body(G, T, G1).
trans_res_S1a(H, T, H1) :- resource_head(H), !,
	add_time(H, T, H1).
trans_res_S1a(S, _, _) :- 
	tllpc_message([resource,'(',S,')',is,invalid]),
	fail.

trans_res_S2(R, _, _) :- var(R), !,
	tllpc_message([resource,'(',R,')',must,be,instantiated]),
	fail.	
trans_res_S2(R, T+N, R1) :- N =:= 0, !,
	trans_res_S2a(R, T, R1).
trans_res_S2(R, T, R1) :- 
	trans_res_S2a(R, T, R1).

trans_res_S2a(R, _, _) :- var(R), !,
	tllpc_message([resource,'(',R,')',must,be,instantiated]),
	fail.	
trans_res_S2a((R1&R2), T, (RR1&RR2)) :- !,
	trans_res_S2a(R1, T, RR1),
	trans_res_S2a(R2, T, RR2).
trans_res_S2a((forall Xs\ S), T, (forall Xs\ SS)) :- !,
	trans_res_S2a(S, T, SS).
trans_res_S2a((G -<> H), T, (forall T1\ (T1 >= T, G1) -<> H1)) :- resource_head(H), !,
	add_time(H, T1, H1),
	trans_body(G, T1, G1).
trans_res_S2a(H, T, (forall T1\ (T1 >= T) -<> H1)) :- resource_head(H), !,
	add_time(H, T1, H1).
trans_res_S2a(S, _, _) :- 
	tllpc_message([resource,'(',S,')',is,invalid]),
	fail.

trans_res_S3(S, _, _) :- var(S),!,
	tllpc_message([resource,'(',S,')',must,be,instantiated]),
	fail.
trans_res_S3((forall Xs\ S), T, (forall Xs\ SS)) :- !,
	trans_res_S3(S, T, SS).
trans_res_S3((G -<> H), _T, (forall T1\ G1 -<> H1)) :- resource_head(H), !,
	add_time(H, T1, H1),
	trans_body(G, T1, G1).
trans_res_S3(H, _T, (forall T1\ H1)) :- resource_head(H), !,
	add_time(H, T1, H1).
trans_res_S3(S, _, _) :- 
	tllpc_message([resource,'(',S,')',is,invalid]),
	fail.
	
/**************************************************************
	Utilities
**************************************************************/

add_time(_, T, _) :- nonvar(T), !,
	tllpc_message([time,T, must,be,unbound,variable]),
	fail.	
add_time(X, T, Y) :-
	X =.. [L|Ls],
	Y =.. [L|[T|Ls]].

tllpc_message([]) :- nl.
tllpc_message([M|Ms]) :-
	write(M), write(' '),
	tllpc_message(Ms).

count_next(@X, N, X1) :- !,
	count_next(X, N1, X1),
	N is N1 + 1.
count_next(X, 0, X).

resource_head(H) :-
	\+ (H = (_ & _)),
	\+ (H = (_ -<> _)),
	\+ (H = (_ :- _)),
	\+ (H = (_ => _)),
	\+ (H = (_ <= _)),
	\+ (H = (forall _ \ _)),
	\+ (H = !(_)),
	\+ (H = @(_)),
	\+ (H = #(_)),
	\+ (H = (_ , _)),
	\+ (H = (_ ; _)).	

